import sys
import os
import logging
import json
import tempfile
import shutil
from . import utils as lib_utils
from . import repo
from . import code_secrets as lib_code_secrets
from . import sast
from . import iac

def get_inventory(args):
    ret_assets = []
    if args.projects == None or args.projects == '':
        logging.error("No projects listed")
        return ret_assets
    plist = args.projects.split(',')
    for proj in plist:
        cmd = "gsutil ls -p "+proj
        out = lib_utils.run_cmd_on_host(args, None, cmd)
        if out == None:
            logging.warning("No cloud storage urls found for "+proj)
            continue
        for gsurl in out.splitlines():
            if 'gcf-sources' not in gsurl:
                logging.warning("No Google Function storage urls found for "+proj)
                continue
            cmd = "gsutil ls -r "+gsurl.strip()
            urlls = lib_utils.run_cmd_on_host(args, None, cmd)
            furldict = {} 
            for l in urlls.splitlines():
                if not l.endswith('function-source.zip'):
                    continue
                fname = l.split('/')[3]
                if fname not in furldict:
                    furldict[fname] = l
                else:
                    efurl = furldict[fname]
                    efver = efurl.split('/')[4].split('-')[1]
                    fver = l.split('/')[4].split('-')[1]
                    if int(fver) > int(efver):
                        furldict[fname] = l
            for f in furldict:
                l = furldict[f]
                zfname = l.split('/')[5]
                fname = l.split('/')[3]
                fver = l.split('/')[4]
                fid = fname
                fguid = fname.split('-')[-5:]
                fguid = '-'.join(fguid)
                fname = fname.replace('-'+fguid,'')
                args.assetid = fid
                args.assetname = fname + ' ' + fver 
                temp_dir = tempfile.mkdtemp()
                args.repo = temp_dir 
                cmd = 'gsutil cp '+l+' '+temp_dir
                out = lib_utils.run_cmd_on_host(args, None, cmd)
                if out == None:
                    logging.error("Error running gsutil command")
                    continue
                cmd = 'cd '+temp_dir+';unzip '+zfname
                out = lib_utils.run_cmd_on_host(args, None, cmd)
                if out == None:
                    logging.error("Error running gsutil command")
                    continue
                logging.info("Discovering Google Function - "+fname+" "+fver)
                logging.debug("Storage url "+l)
                logging.info("Discovering code dependencies for vulnerability scan.")
                assets = repo.discover_inventory(args, temp_dir)
                if args.secrets_scan:
                    logging.info("Discovering secrets/sensitive information. This may take some time.")
                    secret_records = lib_code_secrets.scan_for_secrets(args, temp_dir, temp_dir)
                    assets[0]['secrets'] = secret_records

                code_issues = []
                if args.sast:
                    logging.info("Performing static analysis. This may take some time.")
                    sast_records = sast.run_sast(args, temp_dir, temp_dir)
                    code_issues.extend(sast_records)

                if args.iac_checks:
                    logging.info("Identifying infrastructure as code (IaC) issues. This may take some time.")
                    iac_records = iac.run_iac_checks(args, temp_dir, temp_dir)
                    code_issues.extend(iac_records)

                if len(code_issues) > 0:
                    assets[0]['sast'] = code_issues
                assets[0]['type'] = 'Google Cloud Function'
                assets[0]['tags'].extend(['GCP', 'Google Function', 'Serverless'])
                ret_assets.extend(assets)
                shutil.rmtree(temp_dir)
    return ret_assets 



import os
import logging

from . import sbom_cyclonedx
from . import sbom_spdx
from . import sbom_tw

SUPPORTED_SBOM_FORMATS_FOR_STANDARD = {
        "cyclonedx": ["json"],
        "spdx": ["tagvalue"],
        "threatworx": ["json", "csv"] # keep this one last
}
SUPPORTED_SBOM_STANDARDS = list(SUPPORTED_SBOM_FORMATS_FOR_STANDARD.keys())

def get_inventory(args):
    assets = []
    sbom_standard = args.standard
    sbom_format = args.format
    sbom_abs_path = os.path.abspath(args.input)

    if sbom_standard not in SUPPORTED_SBOM_STANDARDS:
        logging.error("Unsupported SBOM standard [%s]", sbom_standard)
        return assets

    supported_sbom_formats = SUPPORTED_SBOM_FORMATS_FOR_STANDARD[sbom_standard]
    if sbom_format not in supported_sbom_formats:
        logging.error("Unsupported format [%s] for SBOM standard [%s]", sbom_format, sbom_standard)
        return assets
    
    if os.path.isfile(sbom_abs_path) == False:
        logging.error("Unable to access SBOM file [%s]", sbom_abs_path)
        return assets

    logging.info("Processing SBOM artifact [%s]", sbom_abs_path)
    if sbom_standard == "cyclonedx":
        if sbom_format == "json":
            assets = sbom_cyclonedx.process_json(sbom_abs_path, args)
    elif sbom_standard == "spdx":
        if sbom_format == "tagvalue":
            assets = sbom_spdx.process_tagvalue(sbom_abs_path, args)
    elif sbom_standard == "threatworx":
        if sbom_format == "json":
            assets = sbom_tw.process_json(sbom_abs_path, args)
        elif sbom_format == "csv":
            assets = sbom_tw.process_csv(sbom_abs_path, args)
    logging.info("Done processing SBOM artifact")

    return assets



import os
import json
import logging
import sys
from . import az_cis_utils as az_cis_utils

def check50(subid):
    logging.info("Processing 50...")
    chk51=""
    chk52=""
    chk53=('No Alert found for policyAssignments</li>\n')
    chk54=('No Alert found for Create or Update NSG</li>\n')
    chk55=('No Alert found for Delete NSG</li>\n')
    chk56=('No Alert found for Create or Update NSG security rule</li>\n')
    chk57=('No Alert found for Delete NSG security rule</li>\n')
    chk58=('No Alert found for Create or Update Security Solution</li>\n')
    chk59=('No Alert found for Delete Security Solution</li>\n')
    chk510=('No Alert found for Create or Update SQL Server FW rules</li>\n')
    chk511=('No Alert found for Delete SQL Server FW rules</li>\n')
    chk512=('No Alert found for Create or Update Security Policy</li>\n')
    chk513=''

    score51=['<font color="red">Failed</font>',0]
    score52=['<font color="red">Failed</font>',0]
    score53=['<font color="red">Failed</font>',0]
    score54=['<font color="red">Failed</font>',0]
    score55=['<font color="red">Failed</font>',0]
    score56=['<font color="red">Failed</font>',0]
    score57=['<font color="red">Failed</font>',0]
    score58=['<font color="red">Failed</font>',0]
    score59=['<font color="red">Failed</font>',0]
    score510=['<font color="red">Failed</font>',0]
    score511=['<font color="red">Failed</font>',0]
    score512=['<font color="red">Failed</font>',0]
    score513=['<font color="red">Failed</font>',0]
    try:
        query51='az monitor log-profiles list --query [*].[id,name,retentionPolicy.days]'
        json_cis51=az_cis_utils.query_az(query51)
        #iteration through log profiles
        if (len(json_cis51)>0):
            for i in range(len(json_cis51)):
                chk51=chk51+('logprofiles: <b><font color="blue">%s</b></font></li></br>\n' % json_cis51[i][1])
                score51=['<font color="green">Passed</font>',1]
                chk52=chk52+('Retendtion Days <b><font color="blue">%d</b></font> for logprofiles: <b><font color="blue">%s</b></font></li></br>\n' % (json_cis51[i][2],json_cis51[i][1]))
                if (json_cis51[i][2]>=365):
                    score52=['<font color="green">Passed</font>',1]
        else:
            chk51="No Log Profile found</li>\n"
            chk52="No Log Profile found</li>\n"
        query53='az monitor activity-log alert list --query [*].[name,condition.allOf]'
        #query53=('az monitor activity-log alert list --query "[?contains(id,\'%s\')].[name,condition.allOf]"' % subid)
        json_cis53=az_cis_utils.query_az(query53)
        # iteration through Alerts
        for j in range(len(json_cis53)):
            if ("Microsoft.Authorization/policyAssignments/write" in json_cis53[j][1][1]['equals']):
                chk53=('Alert Name:<font color="blue"><b>%s</b></font> set up for policyAssignments</li>\n' % (json_cis53[j][0]))
                score53=['<font color="green">Passed</font>',1]
            elif ("Microsoft.Network/networkSecurityGroups/write" in json_cis53[j][1][1]['equals']):
                chk54=('Alert Name:<font color="blue"><b>%s</b></font> set up for Create or Update NSG</li>\n' % (json_cis53[j][0]))
                score54=['<font color="green">Passed</font>',1]
            elif ("Microsoft.Network/networkSecurityGroups/delete" in json_cis53[j][1][1]['equals']):
                chk55=('Alert Name:<font color="blue"><b>%s</b></font> set up for Delete NSG</li>\n' % (json_cis53[j][0]))
                score55=['<font color="green">Passed</font>',1]
            elif ("Microsoft.Network/networkSecurityGroups/securityRules/write" in json_cis53[j][1][1]['equals']):
                chk56=('Alert Name:<font color="blue"><b>%s</b></font> set up for Create or Update NSG rule</li>\n' % (json_cis53[j][0]))
                score56=['<font color="green">Passed</font>',1]
            elif ("Microsoft.Network/networkSecurityGroups/securityRules/delete" in json_cis53[j][1][1]['equals']):
                chk57=('Alert Name:<font color="blue"><b>%s</b></font> set up for Delete NSG rule</li>\n' % (json_cis53[j][0]))
                score57=['<font color="green">Passed</font>',1]
            elif ("Microsoft.Security/securitySolutions/write" in json_cis53[j][1][1]['equals']):
                chk58=('Alert Name:<font color="blue"><b>%s</b></font> set up for Create or Update Security Solution</li>\n' % (json_cis53[j][0]))
                score58=['<font color="green">Passed</font>',1]
            elif ("Microsoft.Security/securitySolutions/delete" in json_cis53[j][1][1]['equals']):
                chk59=('Alert Name:<font color="blue"><b>%s</b></font> set up for Delete Security Solution</li>\n' % (json_cis53[j][0]))
                score59=['<font color="green">Passed</font>',1]
            elif ("Microsoft.Sql/servers/firewallRules/write" in json_cis53[j][1][1]['equals']):
                chk510=('Alert Name:<font color="blue"><b>%s</b></font> set up for Create or Update SQL Server FW rules</li>\n' % (json_cis53[j][0]))
                score510=['<font color="green">Passed</font>',1]   
            elif ("Microsoft.Sql/servers/firewallRules/delete" in json_cis53[j][1][1]['equals']):
                chk511=('Alert Name:<font color="blue"><b>%s</b></font> set up for Delete SQL Server FW rules</li>\n' % (json_cis53[j][0]))
                score511=['<font color="green">Passed</font>',1]
            elif ("Microsoft.Security/policies/write" in json_cis53[j][1][1]['equals']):
                chk512=('Alert Name:<font color="blue"><b>%s</b></font> set up for Create or Update Security Policy</li>\n' % (json_cis53[j][0]))
                score512=['<font color="green">Passed</font>',1]
        try:
            query513=('az keyvault list --query [*].[name,id]')
            #query513=('az keyvault list --query "[?contains(id,\'%s\')].[name,id]"' % subid)
            json_cis513=az_cis_utils.query_az(query513)
            if (len(json_cis513)>0):
                # iteration through keyvault
                for j in range(len(json_cis513)):
                    query5131=('az monitor diagnostic-settings list --resource %s' % (json_cis513[j][1]))
                    json_cis5131=az_cis_utils.query_az(query5131)
                    if (len(json_cis5131['value'])>0):
                        category = json_cis5131['value'][0]['logs'][0]['category']
                        days = json_cis5131['value'][0]['logs'][0]['retentionPolicy']['days']
                        if ("AuditEvent" in category):
                            chk513=chk513+('Keyvault <b>%s</b> Audit event enabled for <b>%s</b> days' % (json_cis513[j][0],days))
                            if (days>=180):
                                score513=['<font color="green">Passed</font>',1]
                            else:
                                score513=['<font color="red">Failed</font>',0]
                        else:
                            score513=['<font color="red">Failed</font>',0]
                            chk513=chk513+('Keyvault <b>%s</b> Audit event disabled' % (json_cis513[j][0]))
                    else:
                        chk513=chk513+('Diagnostics disable for Keyvault <b>%s</b>' % (json_cis513[j][0]))
                        score513=['<font color="red">Failed</font>',0]
            else:
                chk513="No KeyVault Found"
                score513=['<font color="green">Passed</font>',1]
        except Exception as e:
            logging.error("Exception in check50: %s %s" %(type(e), str(e.args)))
            unkScore=['<font color="orange">UNKNOWN </font>',0]
            chk="Failed query KeyVault"
            return [chk,chk,chk,chk,chk,chk,chk,chk,chk,chk,chk,chk,chk,unkScore,unkScore,unkScore,unkScore,unkScore,unkScore,unkScore,unkScore,unkScore,unkScore,unkScore,unkScore,unkScore]
            #chk513="Failed query KeyVault"
        return [chk51,chk52,chk53,chk54,chk55,chk56,chk57,chk58,chk59,chk510,chk511,chk512,chk513,score51,score52,score53,score54,score55,score56,score57,score58,score59,score510,score511,score512,score513]
    except Exception as e:
        logging.error("Exception in check50: %s %s" %(type(e), str(e.args)))
        unkScore=['<font color="orange">UNKNOWN </font>',0]
        chk="Failed to query log profiles"
        return [chk,chk,chk,chk,chk,chk,chk,chk,chk,chk,chk,chk,chk,unkScore,unkScore,unkScore,unkScore,unkScore,unkScore,unkScore,unkScore,unkScore,unkScore,unkScore,unkScore,unkScore] 



import sys
import os
import logging
from .gcp_cis_tool import gcp_cis as gcp_cis_tool
from . import utils

def run_gcp_cis_bench(args):
    asset = { }
    if args.assetid.strip() == "":
        logging.error("[assetid] cannot be empty")
        utils.tw_exit(1)
    asset['id'] = args.assetid
    asset['name'] = args.assetname if args.assetname and args.assetname.strip() != "" else asset['id']
    asset['type'] = 'Google Cloud Platform'
    asset['owner'] = args.handle
    asset['products'] = []
    asset['tags'] = ['Google Cloud Platform', 'CIS']
    asset['config_issues'] = gcp_cis_tool.run_tests(args)
    args.no_scan = True
    return asset

def get_inventory(args):
    asset = run_gcp_cis_bench(args)
    return [ asset ]



import sys
import logging
from . import gcp_cis_utils as gcp_cis_utils

def check3_1():
    # 3.1 Ensure that the default network does not exist in a project (Scored)

    logging.info("3.1 Ensure that the default network does not exist in a project (Scored)")

    details = []
    projects = gcp_cis_utils.get_compute_enabled_projects()
    for p in projects:
        out_json = gcp_cis_utils.run_gcloud_cmd("compute networks list --project=%s" % p)
        for entry in out_json:
            if entry.get('name') == 'default':
                details.append(("Default network exists for project [%s]" % p, [p], p))
    if len(details) > 0:
        return gcp_cis_utils.create_issue('cis-gcp-bench-check-3.1', '3.1 [Level 2] Ensure that the default network does not exist in a project (Scored)', details, '5', '', '')
    return None

def check3_2():
    # 3.2 Ensure legacy networks do not exist for a project (Scored)

    logging.info("3.2 Ensure legacy networks do not exist for a project (Scored)")

    details = []
    projects = gcp_cis_utils.get_compute_enabled_projects()
    for p in projects:
        out_json = gcp_cis_utils.run_gcloud_cmd("compute networks list --project=%s" % p)
        for entry in out_json:
            if entry.get('x_gcloud_subnet_mode') == 'LEGACY':
                details.append(("Legacy network [%s] exists for project [%s]" % (entry['name'], p), [entry['name'], p], p))
    if len(details) > 0:
        return gcp_cis_utils.create_issue('cis-gcp-bench-check-3.2', '3.2 [Level 1] Ensure legacy networks do not exist for a project (Scored)', details, '4', '', '')
    return None

def check3_3():
    # 3.3 Ensure that DNSSEC is enabled for Cloud DNS (Scored)

    logging.info("3.3 Ensure that DNSSEC is enabled for Cloud DNS (Scored)")

    details = []
    projects = gcp_cis_utils.get_dns_enabled_projects()
    for p in projects:
        out_json = gcp_cis_utils.run_gcloud_cmd("dns managed-zones list --project=%s" % p)
        for entry in out_json:
            if entry.get('visibility') == 'public':
                dnssec = entry.get('dnssecConfig')
                if dnssec is None or dnssec.get('state') != 'on':
                    details.append(("DNSSEC is not enabled for Managed Zone [%s] in  project [%s]" % (entry['name'], p), [entry['name'], p], p))
    if len(details) > 0:
        return gcp_cis_utils.create_issue('cis-gcp-bench-check-3.3', '3.3 [Level 1] Ensure that DNSSEC is enabled for Cloud DNS (Scored)', details, '4', '', '')
    return None

def _check_DNSSEC_key_setting(keyType, details_msg):
    details = []
    projects = gcp_cis_utils.get_dns_enabled_projects()
    for p in projects:
        out_json = gcp_cis_utils.run_gcloud_cmd("dns managed-zones list --project=%s" % p)
        for entry in out_json:
            if entry.get('visibility') == 'public':
                dnssec = entry.get('dnssecConfig')
                if dnssec is not None:
                    defaultKeySpecs = dnssec.get('defaultKeySpecs')
                    if defaultKeySpecs is not None:
                        for entry_2 in defaultKeySpecs:
                            if entry_2['keyType'] == keyType and entry_2['algorithm'].lower() == "rsasha1":
                                details.append((details_msg % (entry['name'], p), [entry['name'], keyType, "rsasha1", p], p))
    return details

def check3_4():
    # 3.4 Ensure that RSASHA1 is not used for the key-signing key in Cloud DNS DNSSEC (Not Scored)

    logging.info("3.4 Ensure that RSASHA1 is not used for the key-signing key in Cloud DNS DNSSEC (Not Scored)")
    details_msg = "RSASHA1 is used for key-signing in Cloud DNS DNSSEC for managed zone [%s] in project [%s]"
    details = _check_DNSSEC_key_setting("keySigning", details_msg)
    if len(details) > 0:
        return gcp_cis_utils.create_issue('cis-gcp-bench-check-3.4', '3.4 [Level 1] Ensure that RSASHA1 is not used for the key-signing key in Cloud DNS DNSSEC (Not Scored)', details, '4', '', '')
    return None

def check3_5():
    # 3.5 Ensure that RSASHA1 is not used for the zone-signing key in Cloud DNS DNSSEC (Not Scored)

    logging.info("3.5 Ensure that RSASHA1 is not used for the zone-signing key in Cloud DNS DNSSEC (Not Scored)")

    details_msg = "RSASHA1 is used for zone-signing in Cloud DNS DNSSEC for managed zone [%s] in project [%s]"
    details = _check_DNSSEC_key_setting("zoneSigning", details_msg)
    if len(details) > 0:
        return gcp_cis_utils.create_issue('cis-gcp-bench-check-3.5', '3.5 [Level 1] Ensure that RSASHA1 is not used for the zone-signing key in Cloud DNS DNSSEC (Not Scored) ', details, '4', '', '')
    return None

def _check_open_port(port_no, details_msg):
    details = []
    projects = gcp_cis_utils.get_compute_enabled_projects()
    for p in projects:
        out_json = gcp_cis_utils.run_gcloud_cmd("compute firewall-rules list --project=%s" % p)
        for entry in out_json:
            if entry['direction'] != 'INGRESS':
                continue
            if entry.get('allowed') is None:
                continue
            for allowed in entry['allowed']:
                port_open = False
                if allowed.get('IPProtocol') is not None and (allowed['IPProtocol'] == 'tcp' or allowed['IPProtocol'] == 'all'):
                    if entry.get('sourceRanges') is None:
                        continue
                    for sr in entry['sourceRanges']:
                        if sr == "0.0.0.0/0":
                            ports = allowed.get('ports')
                            if ports is None:
                                port_open = True
                            else:
                                for port in ports:
                                    if '-' in port:
                                        # range of ports is given
                                        start_port = int(port.split('-')[0])
                                        end_port = int(port.split('-')[1])
                                        if start_port <= port_no and port_no <= end_port:
                                            port_open = True
                                    elif port_no == int(port):
                                        port_open = True
                if port_open:
                    details.append((details_msg % p, [p, str(port_no)], p))
                    return details
    return details

def check3_6():
    # 3.6 Ensure that SSH access is restricted from the internet (Scored)

    logging.info("3.6 Ensure that SSH access is restricted from the internet (Scored)")
    details_msg = "SSH access [port 22] is open from the internet for project [%s]"
    details = _check_open_port(22, details_msg)
    if len(details) > 0:
        return gcp_cis_utils.create_issue('cis-gcp-bench-check-3.6', '3.6 [Level 2] Ensure that SSH access is restricted from the internet (Scored) ', details, '5', '', '')
    return None

def check3_7():
    # 3.7 Ensure that RDP access is restricted from the Internet (Scored)

    logging.info("3.7 Ensure that RDP access is restricted from the Internet (Scored)")
    details_msg = "RDP access [port 3389] is open from the internet for project [%s]"
    details = _check_open_port(3389, details_msg)
    if len(details) > 0:
        return gcp_cis_utils.create_issue('cis-gcp-bench-check-3.7', '3.7 [Level 2] Ensure that RDP access is restricted from the Internet (Scored) ', details, '5', '', '')
    return None

def check3_8():
    # 3.8 Ensure that VPC Flow Logs is enabled for every subnet in a VPC Network (Scored)

    logging.info("3.8 Ensure that VPC Flow Logs is enabled for every subnet in a VPC Network (Scored)")
    details = []
    projects = gcp_cis_utils.get_compute_enabled_projects()
    for p in projects:
        out_json = gcp_cis_utils.run_gcloud_cmd("compute networks subnets list --project=%s" % p)
        for entry in out_json:
            enableFlowLogs = entry.get('enableFlowLogs')
            if enableFlowLogs is None or enableFlowLogs == False:
                details.append(("Subnet [%s] with IP address range [%s] for project [%s] does not have VPC Flow Logs enabled" % (entry['name'], entry['ipCidrRange'], p), [entry['name'], entry['ipCidrRange'], p], entry['name']))
    if len(details) > 0:
        return gcp_cis_utils.create_issue('cis-gcp-bench-check-3.8', '3.8 [Level 1] Ensure that VPC Flow Logs is enabled for every subnet in a VPC Network (Scored)', details, '4', '', '')
    return None

def _check_target_proxy(p, cmd, msg_prefix):
    details = []
    out_json = gcp_cis_utils.run_gcloud_cmd("compute %s list --project=%s" % (cmd, p))
    for entry in out_json:
        sslPolicy = entry.get('sslPolicy')
        if sslPolicy is None:
            details.append(("%s proxy load balancer [%s] in project [%s] uses GCP default policy which is insecure" % (msg_prefix, entry['name'], p), [msg_prefix, entry['name'], p], entry['name']))
        else:
            sslPolicyName = sslPolicy.split('/')[-1]
            out_json_2 = gcp_cis_utils.run_gcloud_cmd("compute ssl-policies describe %s --project=%s" % (sslPolicyName, p))
            policySatisfied = False
            if out_json_2['profile'] == "MODERN" and out_json_2['minTlsVersion'] == "TLS_1_2":
                policySatisfied = True
            if out_json_2['profile'] == "RESTRICTED":
                policySatisfied = True
            if out_json_2['profile'] == "CUSTOM":
                not_allowed_ciphers = set()
                not_allowed_ciphers.add("TLS_RSA_WITH_AES_128_GCM_SHA256")
                not_allowed_ciphers.add("TLS_RSA_WITH_AES_256_GCM_SHA384")
                not_allowed_ciphers.add("TLS_RSA_WITH_AES_128_CBC_SHA")
                not_allowed_ciphers.add("TLS_RSA_WITH_AES_256_CBC_SHA")
                not_allowed_ciphers.add("TLS_RSA_WITH_3DES_EDE_CBC_SHA")
                enabled_ciphers = set(out_json_2['enabledFeatures'])
                weak_ciphers_used = enabled_ciphers.intersection(not_allowed_ciphers)
                if len(weak_ciphers_used) == 0:
                    policySatisfied = True

            if policySatisfied == False:
                details.append(("%s proxy load balancer [%s] in project [%s] uses SSL policy with weak ciphers" % (msg_prefix, entry['name'], p), [msg_prefix, entry['name'], p], entry['name']))
    return details

def check3_9():
    # 3.9 Ensure no HTTPS or SSL proxy load balancers permit SSL policies with weak cipher suites (Not Scored)

    logging.info("3.9 Ensure no HTTPS or SSL proxy load balancers permit SSL policies with weak cipher suites (Not Scored)")
    details = []
    proxy_types = ["target-https-proxies", "target-ssl-proxies"]
    projects = gcp_cis_utils.get_compute_enabled_projects()
    for p in projects:
        details.extend(_check_target_proxy(p, "target-https-proxies", "HTTPS"))
        details.extend(_check_target_proxy(p, "target-ssl-proxies", "SSL"))
    if len(details) > 0:
        return gcp_cis_utils.create_issue('cis-gcp-bench-check-3.9', '3.9 [Level 1] Ensure no HTTPS or SSL proxy load balancers permit SSL policies with weak cipher suites (Not Scored)', details, '4', '', '')
    return None

def run_checks():
    config_issues = []
    gcp_cis_utils.append_issue(config_issues, check3_1())
    gcp_cis_utils.append_issue(config_issues, check3_2())
    gcp_cis_utils.append_issue(config_issues, check3_3())
    gcp_cis_utils.append_issue(config_issues, check3_4())
    gcp_cis_utils.append_issue(config_issues, check3_5())
    gcp_cis_utils.append_issue(config_issues, check3_6())
    gcp_cis_utils.append_issue(config_issues, check3_7())
    gcp_cis_utils.append_issue(config_issues, check3_8())
    gcp_cis_utils.append_issue(config_issues, check3_9())
    return config_issues




import sys
import platform
import os
import logging
import json
from . import utils
from .gcp_cis_tool import gcp_cis_utils

def get_cos_packages(package_list):
    plist = []
    for package in package_list:
        plist.append(package['Arch'] + ' ' + package['Name'] + ' ' + package['Version'])
    return plist

def get_deb_packages(package_list):
    plist = []
    for package in package_list:
        plist.append(package['Name'] + ' ' + package['Version'])
    return plist

def get_rpm_packages(package_list):
    plist = []
    for package in package_list:
        arch = 'noarch' if package['Arch'] == 'all' or package['Arch'] == '(none)' else package['Arch']
        plist.append(package['Name'] + ' ' + package['Version'] + '.' + arch)
    return plist

def get_cos_version(ci_json):
    if len(ci_json['InstalledPackages']['cos']) > 0:
        cosver = ci_json['SystemInformation']['ShortName'] + '-' + ci_json['SystemInformation']['Version'] + '-'
        cosver = cosver + ci_json['InstalledPackages']['cos'][0]['Version']
        cosver = cosver.replace('.','-')
        return cosver
    return None

def get_installed_packages(ci_json):
    plist = []
    ip = ci_json['InstalledPackages']
    for ptype in ip.keys():
        if ptype == "deb":
            plist.extend(get_deb_packages(ip[ptype]))
        elif ptype == "rpm":
            plist.extend(get_rpm_packages(ip[ptype]))
        elif ptype in ['zypperPatches', 'gem']:
            # Skip these packages
            continue
        elif ptype == 'cos':
            plist.extend(get_cos_packages(ip[ptype]))
        else:
            logging.error("Error: Compute instance [%s] has unknown package type [%s]", ci_json['SystemInformation']['LongName'], ptype)
    return plist

def is_compute_running(project, ci_name, ci_zone):
    ci_status_cmd = "compute instances --project=%s describe %s --zone=%s" % (project, ci_name, ci_zone)
    ci_json = gcp_cis_utils.run_gcloud_cmd(ci_status_cmd)
    if ci_json['status'] == 'RUNNING':
        return True
    return False

def process_compute_inventory_json(args, project_id, ci_id, ci_json):
    ci_name = ci_json['SystemInformation']['Hostname']
    asset_os = ci_json['SystemInformation']['LongName']
    logging.info("Collecting inventory for compute instance [%s] with ID [%s]", ci_name, ci_id)
    plist = get_installed_packages(ci_json)

    asset_data = {}
    asset_data['id'] = ci_id
    asset_data['name'] = project_id + '_' + ci_name
    atype = utils.get_asset_type(asset_os)
    asset_data['type'] = atype if atype is not None else 'Other'
    asset_data['owner'] = args.handle
    asset_data['products'] = plist
    asset_tags = []
    asset_tags.append('OS_RELEASE:' + asset_os)
    asset_tags.append('GCP_PROJECT:' + project_id)
    if atype == 'Google Container-Optimized OS':
        cosversion = get_cos_version(ci_json)
        if cosversion:
            asset_tags.append('COS_VERSION:'+cosversion)
    if atype is not None:
        asset_tags.append(atype)
    asset_tags.append('SOURCE:GCP')
    asset_data['tags'] = asset_tags
    return asset_data

def get_inventory(args):
    assets = []
    gcp_cis_utils.set_encoding(args.encoding)
    compute_instances_by_project = gcp_cis_utils.get_compute_instances_with_os_inventory_by_projects()
    for project in compute_instances_by_project.keys():
        logging.info("Processing project [%s]", project)
        compute_instances_json = compute_instances_by_project[project]
        for compute_instance in compute_instances_json:
            ci_id = compute_instance['id']
            ci_name = compute_instance['name']
            ci_zone = compute_instance['zone'].split('/')[-1]

            if is_compute_running(project, ci_name, ci_zone) == False:
                # Do not refresh assets which are not running anymore
                continue

            ci_inventory_cmd = "compute instances os-inventory --project=%s describe %s --zone=%s" % (project, ci_name, ci_zone)
            ci_inventory_json = gcp_cis_utils.run_gcloud_cmd(ci_inventory_cmd)
            asset = process_compute_inventory_json(args, project, ci_id, ci_inventory_json)
            if asset is not None:
                assets.append(asset)
    logging.info("Completed inventory collection...")
    return assets




import json
import logging

from . import sbom_utils

def get_technology(component):
    technology = None
    purl = component.get('purl')
    if purl is None:
        return technology
    tokens = purl.split('/')
    if len(tokens) > 1:
        if tokens[0].startswith("pkg:"):
            technology = tokens[0][4:]
    return technology

def process_json_components(products, shallow_technology_products, technology_products, components, level):
    for component in components:
        ctype = component['type']
        if ctype not in ['library','framework']:
            continue
        cname = component['name']
        cversion = component['version']
        cgroup = component.get('group')
        cscope = component.get('scope')
        if cscope == "excluded":
            continue
        ctech = get_technology(component)
        product = cname + " " + cversion
        if cgroup is not None and len(cgroup) > 0:
            if cgroup[0] == "@":
                product = cgroup + "/" + product
            else:
                product = cgroup + ":" + product
        products.add(product)

        if ctech is not None:
            if level == 1:
                # Add to shallow_technology_products
                tp_set = shallow_technology_products.get(ctech)
                if tp_set is None:
                    tp_set = set()
                    shallow_technology_products[ctech] = tp_set
                tp_set.add(product)

            # Add to technology_products
            tp_set = technology_products.get(ctech)
            if tp_set is None:
                tp_set = set()
                technology_products[ctech] = tp_set
            tp_set.add(product)

        # process sub-components
        sub_components = component.get('components')
        if sub_components is not None and len(sub_components) > 0:
            process_json_components(products, shallow_technology_products, technology_products, sub_components, level + 1)

        # add logic to consume License information

def process_json(sbom_abs_path, args):
    sbom_json = None
    with open(sbom_abs_path, 'rb') as sbom_fd:
        try:
            sbom_json = json.load(sbom_fd)
        except ValueError:
            logging.error("JSON parsing failed for [%s]", sbom_abs_path)
            return  []

    technology_products = { }
    shallow_technology_products = { }
    products = set()
    components = sbom_json.get('components')
    if components is None:
        logging.warning("No components found in SBOM")
        return []

    process_json_components(products, shallow_technology_products, technology_products, components, 1)

    tags  = set()
    tags.add('SBOM')

    # convert shallow_technology_products set to list
    sbom_utils.convert_technology_products(shallow_technology_products, tags)

    # convert technology_products set to list
    sbom_utils.convert_technology_products(technology_products, tags)

    asset_id = sbom_utils.get_asset_id(args)

    asset_data = { }

    asset_data['id'] = asset_id
    if args.assetname is None or len(args.assetname.strip()) == 0:
        asset_data['name'] = asset_data['id']
    else:
        asset_data['name'] = args.assetname
    asset_data['type'] = 'Source Repository'
    asset_data['owner'] = args.handle
    asset_data['products'] = list(products) # convert products set to list
    if len(technology_products) > 0:
        asset_data['compliance_metadata'] = {"source_metadata": {"technology_products": technology_products, "shallow_technology_products": shallow_technology_products}}
    asset_data['tags'] = list(tags)

    return [ asset_data ]




import logging

from . import sbom_utils

def process_tagvalue(sbom_abs_path, args):
    all_products = { }
    products = set()
    main_product_spdxid = None
    direct_dependencies = set()
    lines = []
    with open(sbom_abs_path, 'r') as sbom_fd:
        package_name = None
        package_spdxid = None
        package_version = None
        lines = sbom_fd.readlines()
    for line in lines:
        line = line.strip()
        if line.startswith('PackageName:'):
            if package_name is not None:
                product = package_name if package_version is None else package_name + ' ' + package_version
                products.add(product)
                package_name = None
                package_spdxid = None
                package_version = None
            package_name = line.split(':')[1].strip()
        if line.startswith('SPDXID:'):
            package_spdxid = line.split(':')[1].strip()
        if line.startswith('PackageVersion:'):
            package_version = line.split(':')[1].strip()
        if line.startswith('ExternalRef:'):
            value = line.split(':')[1].strip()
            value = value.split()
            if value[0] == "PACKAGE-MANAGER":
                if value[1] == "maven-central":
                    package_technology = "maven"
                else:
                    package_technology = value[1]
                product_dict = { }
                product_dict['name'] = package_name
                product_dict['version'] = package_version
                product_dict['technology'] = package_technology
                product_dict['spdxid'] = package_spdxid
                all_products[package_spdxid] = product_dict
        if line.startswith('Relationship:'):
            value = line.split(':')[1].strip()
            tokens = value.split()
            if tokens[1] == "DESCRIBES":
                main_product_spdxid = tokens[2]
            elif main_product_spdxid is not None and tokens[1] in ["CONTAINS", "DEPENDS_ON", "DYNAMIC_LINK", "STATIC_LINK", "HAS_PREREQUISITE"] and tokens[0] == main_product_spdxid:
                direct_dependencies.add(tokens[2])
            elif main_product_spdxid is not None and tokens[1] in ["CONTAINED_BY", "DEPENDENCY_OF", "BUILD_DEPENDENCY_OF", "DEV_DEPENDENCY_OF", "RUNTIME_DEPENDENCY_OF", "PREREQUISITE_FOR"] and tokens[2] == main_product_spdxid:
                direct_dependencies.add(tokens[0])
    if package_name is not None:
        product = package_name if package_version is None else package_name + ' ' + package_version
        products.add(product)

    if len(all_products) > 0 or len(direct_dependencies) > 0:
        shallow_technology_products = { }
        for dd in direct_dependencies:
            product = all_products.get(dd)
            if product is not None:
                technology = product['technology']
                pn = product['name'] if product.get('version') is None else product['name'] + ' ' + product['version']
                tp_set = shallow_technology_products.get(technology)
                if tp_set is None:
                    tp_set = set()
                    shallow_technology_products[technology] = tp_set
                tp_set.add(pn)
        technology_products = { }
        for key in all_products:
            product = all_products[key]
            pn = product['name'] if product.get('version') is None else product['name'] + ' ' + product['version']
            technology = product['technology']
            tp_set = technology_products.get(technology)
            if tp_set is None:
                tp_set = set()
                technology_products[technology] = tp_set
            tp_set.add(pn)

    tags = set()
    tags.add('SBOM')

    # convert shallow_technology_products set to list
    sbom_utils.convert_technology_products(shallow_technology_products, tags)

    # convert technology_products set to list
    sbom_utils.convert_technology_products(technology_products, tags)

    asset_id = sbom_utils.get_asset_id(args)

    asset_data = { }

    asset_data['id'] = asset_id
    if args.assetname is None or len(args.assetname.strip()) == 0:
        asset_data['name'] = asset_data['id']
    else:
        asset_data['name'] = args.assetname
    asset_data['type'] = 'Source Repository'
    asset_data['owner'] = args.handle
    asset_data['products'] = list(products) # convert products set to list
    if len(technology_products) > 0:
        asset_data['compliance_metadata'] = {"source_metadata": {"technology_products": technology_products, "shallow_technology_products": shallow_technology_products}}
    asset_data['tags'] = list(tags)

    return [ asset_data ]




import sys
import logging
import json
import csv

from . import utils

all_asset_types = None

def get_assets_from_json_file(in_file):
    assets = []
    with open(in_file,'r') as fd:
        try:
            sbom_json = json.load(fd)
            assets = sbom_json['assets']
        except ValueError:
            logging.error("Error loading JSON file [%s]", in_file)
            utils.tw_exit(1)
    return assets

def get_assets_from_csv_file(in_file, args):
    assets = []
    with open(in_file, 'rU') as fd:
        csv_reader = csv.DictReader(fd, delimiter=',', escapechar='\\')
        for record in csv_reader:
            asset = create_asset_from_csv_record(record, args)
            if asset is not None:
                assets.append(asset)
    return assets

def create_asset_from_csv_record(record, args):
    asset = { }
    asset['id'] = record['Asset Id'].strip()
    asset['name'] = record['Asset Name'].strip()
    asset['type'] = record['Asset Type'].strip()
    asset['owner'] = record['Owner'].strip()
    asset['tags'] = get_mv_field(record['Tags'])
    asset['products'] = get_mv_field(record['Products'])
    asset['patches'] = get_mv_field(record['Patches'])

    if validate_update_csv_asset(asset, args) == False:
        return None

    return asset

def get_mv_field(mv_field_value):
    mv_field_values = mv_field_value.split(';')
    ret_val = []
    for item in mv_field_values:
        item = item.strip()
        if item == '':
            continue
        ret_val.append(item)
    return ret_val

def validate_update_csv_asset(asset, args):
    global all_asset_types
    if asset['id'] == '':
        return False
    if asset['name'] == '':
        return False
    if asset['type'] == '':
        asset['type'] = 'Other'
    else:
        if all_asset_types is None:
            url = "https://" + args.instance + "/api/v1/assets/types"
            auth_data = "?handle=" + args.handle + "&token=" + args.token + "&format=json"
            response = utils.requests_get(url + auth_data)
            if response is not None and response.status_code != 200:
                logging.error("Unable to get valid asset types.")
                utils.tw_exit(1)
            all_asset_types = response.json()
        if asset['type'] not in all_asset_types:
            logging.warning("Skipping asset [%s] with invalid asset type [%s]", asset['name'],asset['type'])
            return False
    if asset['owner'] == '':
        asset['owner'] = args.handle
    return True

def process_json(sbom_abs_path, args):
    assets = get_assets_from_json_file(sbom_abs_path)
    return assets

def process_csv(sbom_abs_path, args):
    if args.token is None or len(args.token) == 0:
        logging.error("API token is required to validate asset types in CSV")
        utils.tw_exit(1)
    assets = get_assets_from_csv_file(sbom_abs_path, args)
    return assets
    



import sys
import re
import os
import shutil
import stat
import subprocess
import logging
import json
import tempfile
import traceback
from .dast_plugins import skipfish as skipfish
from .dast_plugins import arachni as arachni 
from . import utils

dast_plugins = ["skipfish", "arachni"]

def get_inventory(args):
    if args.plugin not in dast_plugins:
        logging.error('Not a supported DAST plugin')
        utils.tw_exit(1) 
    if args.assetid.strip() == "":
        logging.error('[assetid] cannot be empty')
        utils.tw_exit(1)
    asset_id = args.assetid
    if asset_id == None:
        asset_id = args.url
    asset_name = None
    if args.assetname == None or args.assetname.strip() == "":
        asset_name = asset_id 
    else:
        asset_name = args.assetname

    asset_data = {}
    asset_data['id'] = asset_id
    asset_data['name'] = asset_name
    asset_data['type'] = "Web Application" 
    asset_data['owner'] = args.handle
    asset_data['products'] = [] 
    asset_data['tags'] = [args.url] 

    logging.info("Running DAST plugin for web application")

    asset_data['config_issues'] = run_dast(args)

    # disable scan
    args.no_scan = True
    return [ asset_data ]

def run_dast(args):
    findings = []
    if args.plugin == 'arachni':
        findings = arachni.run(args)
    elif args.plugin == 'skipfish':
        findings = skipfish.run(args)
    #print json.dumps(findings, indent=4)
    return findings



import sys
import os
import subprocess
import logging
import json
import yaml
import uuid
import tempfile

from . import utils as lib_utils
from . import docker

def get_helm_command_path():
    HELM_CMD = os.environ.get('HELM_PATH')
    if HELM_CMD is None:
        HELM_CMD = "/usr/local/bin/helm"
    else:
        if os.path.isfile(HELM_CMD) == False:
            logging.error("Helm command not found at specified HELM_PATH [%s]", HELM_CMD)
            lib_utils.tw_exit(1)
        elif os.access(HELM_CMD, os.X_OK) == False:
            logging.error("Helm command file [%s] is not an executable", HELM_CMD)
            lib_utils.tw_exit(1)
    return HELM_CMD

def get_deployment_name(yaml_json):
    if yaml_json.get('metadata') is None:
        return None
    namespace = yaml_json['metadata'].get('namespace')
    deployment_name = yaml_json['metadata'].get('name')
    return namespace, deployment_name

def get_list_of_containers(yaml_json):
    kind = yaml_json.get('kind')
    init_container_list = []
    container_list = []
    if yaml_json.get('spec') is None:
        return container_list

    if kind == "Pod":
        if yaml_json['spec'].get('containers') is None:
            return init_container_list, container_list
        yaml_containers = yaml_json['spec']['containers']
        yaml_init_containers = yaml_json['spec'].get('initContainers')
    elif kind == "CronJob":
        if yaml_json['spec'].get('jobTemplate') is None:
            return init_container_list, container_list
        if yaml_json['spec']['jobTemplate'].get('spec') is None:
            return init_container_list, container_list
        if yaml_json['spec']['jobTemplate']['spec'].get('template') is None:
            return init_container_list, container_list
        if yaml_json['spec']['jobTemplate']['spec']['template'].get('spec') is None:
            return init_container_list, container_list
        if yaml_json['spec']['jobTemplate']['spec']['template']['spec'].get('containers') is None:
            return init_container_list, container_list
        yaml_containers = yaml_json['spec']['jobTemplate']['spec']['template']['spec']['containers']
        yaml_init_containers = yaml_json['spec']['jobTemplate']['spec']['template']['spec'].get('initContainers')
    else:
        if yaml_json['spec'].get('template') is None:
            return init_container_list, container_list
        if yaml_json['spec']['template'].get('spec') is None:
            return init_container_list, container_list
        if yaml_json['spec']['template']['spec'].get('containers') is None:
            return init_container_list, container_list
        yaml_containers = yaml_json['spec']['template']['spec']['containers']
        yaml_init_containers = yaml_json['spec']['template']['spec'].get('initContainers')

    if yaml_init_containers is not None:
        for yaml_init_container in yaml_init_containers:
            yaml_init_container_image = yaml_init_container.get('image')
            if yaml_init_container_image is not None:
                init_container_list.append(yaml_init_container_image)

    for yaml_container in yaml_containers:
        yaml_container_image = yaml_container.get('image')
        if yaml_container_image is not None:
            container_list.append(yaml_container_image)

    return init_container_list, container_list

def discover_containers(args, container_list, asset_name, namespace, deployment_name, kind, extra_tags = []):
    allassets = []
    if len(container_list) == 0:
        return allassets;
    for container in container_list:
        logging.info("Discovering Kubernetes container: %s", container)
        digest = container.split('@')[1] if '@' in container else None
        args.image = container
        args.assetid = asset_name + '-' + container
        args.assetid = args.assetid.replace('/','-')
        args.assetid = args.assetid.replace(':','-')
        args.assetname = asset_name + '-' + container
        logging.info("Discovering image "+args.image)
        assets = docker.get_inventory(args, digest)
        if assets:
            for a in assets:
                if a.get('tags') is None:
                    a['tags'] = []
                a['tags'].append('KubernetesCluster')
                if namespace is not None:
                    a['tags'].append(namespace)
                a['tags'].append(deployment_name)
                a['tags'].append(kind)
                for extra_tag in extra_tags:
                    a['tags'].append(extra_tag)
            allassets = allassets + assets
    return allassets

def discover_assets_from_yaml(args, k8s_yaml, asset_name_override=None):
    allassets = [] 
    with open(k8s_yaml, 'r') as yaml_fd:
        yaml_jsons = yaml.load_all(yaml_fd, Loader=yaml.FullLoader)
        if yaml_jsons is None:
            logging.error("Unable to load YAML")
            return allassets
        for yaml_json in yaml_jsons:
            if yaml_json is None:
                continue
            kind = yaml_json.get('kind')
            if kind is None or kind == "" or kind not in ["Deployment", "ReplicaSet", "Pod", "PodTemplate", "StatefulSet", "DaemonSet", "Job", "CronJob", "ReplicationController"]:
                # only process those 'kind' with container image
                continue
            namespace, deployment_name = get_deployment_name(yaml_json)
            extra_tags = []
            if asset_name_override is not None:
                asset_name = asset_name_override
                extra_tags = ["HelmChart"]
            else:
                asset_name = deployment_name if namespace is None else namespace + '-' + deployment_name
            logging.info("Processing Kubernetes workload resource: %s", kind)
            init_container_list, container_list = get_list_of_containers(yaml_json)
            if len(init_container_list) == 0 and len(container_list) == 0:
                logging.warning("No containers referenced in %s YAML", kind)
                continue
            allassets = allassets + discover_containers(args, container_list, asset_name, namespace, deployment_name, kind, extra_tags)
            extra_tags.append("InitContainer")
            allassets = allassets + discover_containers(args, init_container_list, asset_name, namespace, deployment_name, kind, extra_tags)

    return allassets

def run_helm_command(cmdarr, encoding):
    try:
        logging.debug("Running command %s", cmdarr)
        out = subprocess.check_output(cmdarr, shell=True)
        return out.decode(encoding)
    except subprocess.CalledProcessError as e:
        logging.error("Error running Helm command")
        logging.debug("[helm] command: %s", cmdarr[0])
        logging.debug("Output of [helm] command: %s", e.output)
        return None

def discover_assets_from_helm_chart(args, helm_chart):
    allassets = []
    hc_name = None
    hc_version = None
    hc_type = None
    temp_template = uuid.uuid4().hex
    temp_template = tempfile.gettempdir() + os.path.sep + temp_template + ".yaml"
    if os.path.isfile(temp_template):
        os.remove(temp_template)
    HELM_CMD = get_helm_command_path()
    cmdarr = [ HELM_CMD + " show chart " + helm_chart + " > " + temp_template ]
    cmd_output = run_helm_command(cmdarr, args.encoding)
    if cmd_output is None:
        logging.error("Unable to get Chart.yaml")
        return allassets
    with open(temp_template, 'r') as chart_yaml_fd:
        yaml_json = yaml.load(chart_yaml_fd, Loader=yaml.FullLoader)
        if yaml_json is None:
            logging.error("Unable to load Chart.yaml")
            return allassets
        hc_name = yaml_json['name']
        hc_version = yaml_json['version']
        hc_type = yaml_json.get('type')
        if hc_type is None:
            hc_type = 'applicaton'
    asset_name_override = hc_name + ':' + hc_version
    if hc_type == 'library':
        logging.warning("Helm Chart [%s] is of library type...Skipping it", asset_name_override)
        return allassets
    logging.info("Processing Helm Chart [%s]", asset_name_override)
    if os.path.isfile(temp_template):
        os.remove(temp_template)
    cmdarr = [ HELM_CMD + " template " + helm_chart + " > " + temp_template ]
    cmd_output = run_helm_command(cmdarr, args.encoding)
    if cmd_output is None:
        logging.error("Unable to get Helm chart template YAML")
        os.remove(temp_template)
        return allassets
    allassets = discover_assets_from_yaml(args, temp_template, asset_name_override)
    if os.path.isfile(temp_template):
        os.remove(temp_template)

    logging.info("Done processing Helm Chart [%s]", asset_name_override)
    return allassets

def get_inventory(args):
    if args.deployment_yaml:
        return discover_assets_from_yaml(args, args.deployment_yaml)
    elif args.helm_chart:
        return discover_assets_from_helm_chart(args, args.helm_chart)
    return []



import sys
import re
import os
import shutil
import stat
import subprocess
import logging
import json
import tempfile
import traceback

issue_list = {
  "10101": "SSL certificate issuer information",
  "10201": "New HTTP cookie added",
  "10202": "New 'Server' header value seen",
  "10203": "New 'Via' header value seen",
  "10204": "New 'X-*' header value seen",
  "10205": "New 404 signature seen",

  "10401": "Resource not directly accessible",
  "10402": "HTTP authentication required",
  "10403": "Server error triggered",
  "10404": "Directory listing enabled",
  "10405": "Hidden files / directories",

  "10501": "All external links",
  "10502": "External URL redirector",
  "10503": "All e-mail addresses",
  "10504": "Links to unknown protocols",
  "10505": "Unknown form field (can't autocomplete)",
  "10601": "HTML form (not classified otherwise)",
  "10602": "Password entry form - consider brute-force",
  "10603": "File upload form",
  "10701": "User-supplied link rendered on a page",
  "10801": "Incorrect or missing MIME type (low risk)",
  "10802": "Generic MIME used (low risk)",
  "10803": "Incorrect or missing charset (low risk)",
  "10804": "Conflicting MIME / charset info (low risk)",
  "10901": "Numerical filename - consider enumerating",
  "10902": "OGNL-like parameter behavior",
  "10909": "Signature match (informational)",

  "20101": "Resource fetch failed",
  "20102": "Limits exceeded, fetch suppressed",
  "20201": "Directory behavior checks failed (no brute force)",
  "20202": "Parent behavior checks failed (no brute force)",
  "20203": "IPS filtering enabled",
  "20204": "IPS filtering disabled again",
  "20205": "Response varies randomly, skipping checks",
  "20301": "Node should be a directory, detection error?",

  "30101": "HTTP credentials seen in URLs",
  "30201": "SSL certificate expired or not yet valid",
  "30202": "Self-signed SSL certificate",
  "30203": "SSL certificate host name mismatch",
  "30204": "No SSL certificate data found",
  "30205": "Weak SSL cipher negotiated",
  "30206": "Host name length mismatch (name string has null byte)",
  "30301": "Directory listing restrictions bypassed",
  "30401": "Redirection to attacker-supplied URLs",
  "30402": "Attacker-supplied URLs in embedded content (lower risk)",
  "30501": "External content embedded on a page (lower risk)",
  "30502": "Mixed content embedded on a page (lower risk)",
  "30503": "HTTPS form submitting to a HTTP URL",
  "30601": "HTML form with no apparent XSRF protection",
  "30602": "JSON response with no apparent XSSI protection",
  "30603": "Auth form leaks credentials via HTTP GET",
  "30701": "Incorrect caching directives (lower risk)",
  "30801": "User-controlled response prefix (BOM / plugin attacks)",
  "30901": "HTTP header injection vector",
  "30909": "Signature match detected",

  "40101": "XSS vector in document body",
  "40102": "XSS vector via arbitrary URLs",
  "40103": "HTTP response header splitting",
  "40104": "Attacker-supplied URLs in embedded content (higher risk)",
  "40105": "XSS vector via injected HTML tag attribute",
  "40201": "External content embedded on a page (higher risk)",
  "40202": "Mixed content embedded on a page (higher risk)",
  "40301": "Incorrect or missing MIME type (higher risk)",
  "40302": "Generic MIME type (higher risk)",
  "40304": "Incorrect or missing charset (higher risk)",
  "40305": "Conflicting MIME / charset info (higher risk)",
  "40401": "Interesting file",
  "40402": "Interesting server message",
  "40501": "Directory traversal / file inclusion possible",
  "40601": "Incorrect caching directives (higher risk)",
  "40701": "Password form submits from or to non-HTTPS page",
  "40909": "Signature match detected (higher risk)",

  "50101": "Server-side XML injection vector",
  "50102": "Shell injection vector",
  "50103": "Query injection vector",
  "50104": "Format string vector",
  "50105": "Integer overflow vector",
  "50106": "File inclusion",
  "50107": "Remote file inclusion",
  "50201": "SQL query or similar syntax in parameters",
  "50301": "PUT request accepted",
  "50909": "Signature match detected (high risk)"
}



def on_rm_error( func, path, exc_info):
    os.chmod( path, stat.S_IWRITE )
    os.unlink( path )

def get_object_id(pdir):
    oid = None 
    if os.path.exists(pdir+'/request.dat'):
        f = open(pdir+'/request.dat')
        oid = f.read()
        oid = oid.split()[1]
        f.close()
    return oid 

def tw_open(in_file, in_encoding):
    if sys.version_info[0] < 3:
        f = open(in_file)
    else:
        f = open(in_file, encoding=in_encoding)
    return f

def get_payload(pdir, encoding):
    out = ''
    if os.path.exists(pdir+'/request.dat'):
        f = tw_open(pdir+'/request.dat', encoding)
        out = out + '\nREQUEST\n'+f.read()
        f.close()
    if os.path.exists(pdir+'/response.dat'):
        f = tw_open(pdir+'/response.dat', encoding)
        out = out + '\nRESPONSE\n'+f.read()
        f.close()
    out = re.sub(r'[^\x00-\x7F]+','', out)
    return out

def parse_skipout(skipout, args):
    logging.info("Analyzing results")
    findings = []
    for root, dirs, files in os.walk(skipout):
        for file in files:
            if file == 'issue_index.js':
                issuefile = (os.path.join(root, file))
                f = open(issuefile, 'r')
                contents = f.read()
                contents = contents.replace('var issue = ','').strip()
                contents = contents.replace("'",'"').strip()
                contents = contents.replace("\\",'\\\\').strip()
                contents = contents[:-1]
                iarr = json.loads(contents)
                if len(iarr) > 0:
                    for i in iarr:
                        issue = {}
                        issue['asset_id'] = args.assetid
                        issue['twc_id'] = 'SKIPFISH-' + str(i['type'])
                        issue['twc_title'] = issue_list[str(i['type'])]
                        issue['rating'] = str(i['type'])[0]
                        oid = get_object_id(os.path.join(root, i['dir'])) 
                        if oid == None:
                            oid = str(hash(i['dir']))
                        issue['object_id'] = oid
                        issue['type'] = 'DAST'
                        issue['object_meta'] = ''
                        issue['details'] = get_payload(os.path.join(root, i['dir']), args.encoding)
                        findings.append(issue)
                f.close()
    return findings

def run(args):
    SKIPFISH = "/usr/bin/skipfish"

    if not os.path.isfile(SKIPFISH) or not os.access(SKIPFISH, os.X_OK):
        logging.error('skipfish not found')
        sys.exit(1) 

    logging.warn("Running skipfish plugin. This could take a while")
    path = tempfile.mkdtemp()
    rparams = " -MEU -uv -o "+path 
    logging.warn("Using reporting options: "+rparams)
    logging.warn("Please do not override the reporting options")

    params = args.args
    if params != None:
        params = " " + params + " " + rparams + " " + args.url
    else:
        params = " " + rparams + " " + args.url
    #logging.info("skipfish command line: "+SKIPFISH + " " + params)
    cmdarr = [SKIPFISH + " " + params]
    try:
        out = subprocess.check_output(cmdarr, shell=True)
    except subprocess.CalledProcessError as cpe:
        logging.error("Error running skipfish")
        logging.debug("[skipfish] command: %s", cmdarr[0])
        logging.debug("Output of [skipfish]: %s", cpe.output)
        shutil.rmtree(path, onerror = on_rm_error)
        return None 
    logging.info("skipfish run completed")

    findings = parse_skipout(path, args)
    shutil.rmtree(path, onerror = on_rm_error)

    return findings



import os

# Please use lower case key names based on products
plugin_registry = {
        "log4j_vulns": {"name": "Check for Log4j vulns", "file": "log4j_vulns.sh", "enabled": True},
        "spring4shell_vulns": {"name": "Check for Spring4Shell vulns", "file": "spring4shell_vulns.sh", "enabled": True},
        "poi_vulns": {"name": "Check for Apache POI vulns", "file": "poi_vulns.sh", "enabled": True},
        "tomcat_vulns": {"name": "Check for Apache Tomcat vulns", "file": "tomcat_vulns.sh", "enabled": True},
        "apache_vulns": {"name": "Check for Apache Server vulns", "file": "apache_vulns.sh", "enabled": True},
        "spring_framework_vulns": {"name": "Check for Spring Framework vulns", "file": "spring_framework_vulns.sh", "enabled": True},
}

def get_plugin_dir():
    return os.path.dirname(os.path.abspath(__file__)) + os.sep + "plugins"

def get_plugin_registry():
    return plugin_registry



import sys
import re
import os
import shutil
import stat
import subprocess
import logging
import json
import tempfile
import traceback

from . import iac_meta

checkov_plugin = "/usr/local/bin/checkov"

sevmap = {"LOW": "1","MEDIUM": "3","HIGH": "4","CRITICAL": "5"}

def get_code_snippet(r):
    code_snippet = ''
    if r['code_block'] is None:
        return code_snippet
    for cl in r['code_block']:
        code_snippet = code_snippet + cl[1]
    return code_snippet

def get_refs(r):
    refs = [ ]
    guideline = r.get('guideline')
    if guideline is not None:
        refs.append(guideline)
    return refs

def run_iac_checks(args, path, base_path):
    findings = []
    if not os.path.isfile(checkov_plugin) or not os.access(checkov_plugin, os.X_OK):
        logging.error('IaC security checks CLI - checkov not found')
        return findings

    params = '--output json --directory ' + path
    
    cmdarr = [checkov_plugin + " " + params]
    logging.debug("Running command %s", cmdarr)
    iac_issues = None
    try:
        out = subprocess.check_output(cmdarr, shell=True)
        iac_issues = json.loads(out)
    except subprocess.CalledProcessError as cpe:
        if cpe.returncode == 1 and len(cpe.output) > 0:
            iac_issues = json.loads(cpe.output)
        else:
            logging.error("Error running IaC security checks CLI [checkov]")
            return findings 
    logging.info("IaC security checks CLI [checkov] checks completed")

    # checkov returns list if there are multiple technologies like terraform, kubernetes
    # else it returns dict
    if type(iac_issues) is dict:
        iac_issues = [ iac_issues ]

    for iac_issue in iac_issues:
        failed_results = iac_issue['results'].get('failed_checks') if 'results' in iac_issue else None
        if failed_results is None:
            continue
        for r in failed_results:
            imeta = iac_meta.metadata.get(r['check_id'])
            finding = {}
            finding['issue_id'] = r['check_id']
            if imeta:
                finding['rating'] = sevmap[imeta['severity']]
            else:
                finding['rating'] = '3' # default rating
            finding['filename'] = r['file_path'][1:]
            if args.no_code:
                finding['code_snippet'] = ''
            else:
                finding['code_snippet'] = get_code_snippet(r)
            finding['lineno_start'] = r['file_line_range'][0] if r['file_line_range'][0] is not None else -1
            finding['lineno_end'] = r['file_line_range'][1] if r['file_line_range'][1] is not None else -1
            if imeta:
                finding['description'] = "<h3>" + r['check_name'] + "</h3>" + imeta['description']
            else:
                finding['description'] = r['check_name']
            finding['resource'] = r['resource']
            finding['refs'] = get_refs(r)
            finding['type'] = 'IaC'
            findings.append(finding)

    return findings



import sys
import platform
import os
import subprocess
import logging
import psutil
import ipaddress
import socket
import json
from xml.dom.minidom import parse, parseString
import csv
from . import linux

NMAP = "/usr/bin/nmap"

def nmap_exists():
    return os.path.isfile(NMAP) and os.access(NMAP, os.X_OK)

def discover(args):
    handle = args.handle
    token = args.token
    instance = args.instance

    hosts = args.hosts

def get_wordpress(args):
    if not nmap_exists():
        logging.error('nmap CLI not found')
        return None
    logging.info("Discovering wordpress website on "+args.host)
    assets = []
    word = [NMAP + ' --script http-wordpress-enum '+ args.host]
    try:
        word_out = subprocess.check_output(word, shell=True)
    except subprocess.CalledProcessError:
        logging.error("Error running nmap discovery for wordpress")
        return None 
    plugins = str(word_out).split('plugins')[-1]
    plugins = plugins.split('_http')[0]
    plugins = plugins.split('|')
    products = []
    for p in plugins:
        p = p.strip()
        if p == '':
            continue
        if p == 'themes':
            continue
        if p.startswith('_'):
            p = p.replace('_','')
            p = p.strip()
        if 'Nmap done:' in p:
            p = p.split('\n')[0]
        products.append(p)
    asset_data = {}
    if args.assetid != None:
        asset_data['id'] = args.assetid
    else:
        asset_data['id'] = args.host
    asset_data['name'] = args.host 
    asset_data['type'] = 'WordPress'
    asset_data['owner'] = args.handle
    asset_data['products'] = products
    asset_data['tags'] = ['wordpress']
    assets.append(asset_data)
    return assets

def get_private_ip_cidrs():
    """
    Returns a list of class A, B, or C private IP CIDRs visible on the local host.
    """
    private_cidrs = []
    for interface in psutil.net_if_addrs().values():
        for address in interface:
            if address.family == socket.AF_INET:
                ip_address = ipaddress.IPv4Address(address.address)
                if ip_address.is_private:
                    if ipaddress.IPv4Network(ip_address).is_private:
                        if ip_address.is_private and \
                                (ipaddress.IPv4Address('10.0.0.0') <= ip_address <= ipaddress.IPv4Address('10.255.255.255') or
                                ipaddress.IPv4Address('172.16.0.0') <= ip_address <= ipaddress.IPv4Address('172.31.255.255') or
                                ipaddress.IPv4Address('192.168.0.0') <= ip_address <= ipaddress.IPv4Address('192.168.255.255')):
                            private_cidrs.append(str(ipaddress.IPv4Network(ip_address).with_prefixlen))
    return private_cidrs

def get_inventory(args):
    if not nmap_exists():
        logging.error('nmap CLI not found')
        return None

    if args.hosts == None:
        args.hosts = get_private_ip_cidrs()
    else:
        args.hosts = args.hosts.split()
    assets = []
    for host in args.hosts:
        logging.info("Fingerprinting "+host)
        cmdarr = [NMAP + ' -oX - -sV --script http-wordpress-enum -A -PN -T4 -F '+host]
        try:
            out = subprocess.check_output(cmdarr, shell=True)
            out = out.decode(args.encoding)
        except subprocess.CalledProcessError:
            logging.error("Error running nmap command")
            return None 

        dom = parseString(out)
        hosts = dom.getElementsByTagName("host")
        for h in hosts:
            addr = h.getElementsByTagName("address")[0]
            addr = addr.getAttribute('addr')
            hostname = addr
            harr = h.getElementsByTagName("hostname")
            if harr != None and len(harr) > 0:
                hostname = h.getElementsByTagName("hostname")[0]
                hostname = hostname.getAttribute('name')
                if hostname == 'linux':
                    hostname = addr

            # check for cpes
            cpes = h.getElementsByTagName("cpe")
            products = []
            for c in cpes:
                cstr = c.firstChild.data
                carr = cstr.split(':')
                prodstr = carr[2] + ' ' + carr[3] + ' '
                if len(carr) >= 5:
                    prodstr += carr[4]
                prodstr = prodstr.strip()
                prodstr = prodstr.replace('_',' ')
                if prodstr not in products:
                    products.append(prodstr)
            if 'linux linux kernel' in products:
                ostype = 'Generic Linux'
            elif 'microsoft windows' in products:
                ostype = 'Generic Windows'
            else:
                ostype = 'Unknown'

            # check for wordpress output
            scripts = h.getElementsByTagName("script")
            for s in scripts:
                if s.getAttribute('id') == 'http-wordpress-enum':
                    wpout = s.getAttribute('output')
                    if wpout != None:
                        wplist = wpout.splitlines()
                        for wp in wplist:
                            wp = wp.strip()
                            if wp == '':
                                continue
                            if wp.startswith('Search limited to'):
                                continue
                            if wp == 'plugins':
                                continue
                            if wp == 'themes':
                                continue
                            prodstr = 'wordpress plugin '+wp
                            products.append(prodstr)
            asset_data = {}
            asset_data['id'] = addr 
            asset_data['name'] = hostname 
            asset_data['type'] = 'Other' 
            asset_data['owner'] = args.handle
            asset_data['products'] = products 
            asset_tags = ["DISCOVERY_TYPE:Unauthenticated"]
            asset_data['tags'] = asset_tags
            if args.no_ssh_audit == False:
                ssh_issues = linux.run_ssh_audit(args, addr, addr)
                if len(ssh_issues) != 0:
                    asset_data['tags'].append('SSH Audit')
                asset_data['config_issues'] = ssh_issues
            assets.append(asset_data)        
    return assets




import sys
import re
import os
import shutil
import stat
import subprocess
import logging
import json
import tempfile
import pefile
import glob
import traceback
import requirements
import re
import zipfile
from xml.dom import minidom
import toml
import re
import ast
import textwrap

from . import utils as lib_utils
from . import code_secrets as lib_code_secrets
from . import sast
from . import iac

GIT_PATH = os.environ.get('GIT_PATH')
if GIT_PATH is None:
    if os.name == 'nt':
        GIT_PATH = 'C:\\Program Files\\Git\\cmd\\git.exe'
    else:
        GIT_PATH = '/usr/bin/git'

SUPPORTED_TYPES = ['pip', 'ruby', 'yarn', 'nuget', 'npm', 'maven', 'gradle', 'dll', 'jar', 'cargo', 'go', 'composer']

def cleanse_semver_version(pv):
    pv = pv.replace('"','')
    pv = pv.replace('~','')
    pv = pv.replace('^','')
    pv = pv.replace('<','')
    pv = pv.replace('>','')
    pv = pv.replace('=','')
    temp_tokens = pv.split()
    if len(temp_tokens) >= 2:
        version = temp_tokens[1]
        version = version.replace('X','0')
        version = version.replace('x','0')
        version = version.replace('*','0')
        pv = temp_tokens[0] + ' ' + version
        # Add any remaining tokens following version (like source)
        for token_index in range(2, len(temp_tokens)):
            pv = pv + " " + temp_tokens[token_index]
    return pv

def discover_composer(args, localpath):
    plist = []
    files = lib_utils.find_files(localpath, 'composer.lock')
    if len(files) <= 0:
        files = lib_utils.find_files(localpath, 'composer.json')
    for file_path in files:
        fp = lib_utils.tw_open(file_path, args.encoding)
        if fp == None:
            continue
        contents = fp.read()
        fp.close()
        composer_json = json.loads(contents)
        if file_path.endswith('composer.json'):
            requires = composer_json['require'] if 'require' in composer_json else {}
            for pname in requires.keys():
                pver = requires[pname]
                prod = pname + " " + pver + " source:"+file_path
                plist.append(prod)
        else:
            packages = composer_json['packages']
            for p in packages:
                prod = p['name'] + ' ' + p['version'] + ' source:'+file_path
                plist.append(prod)
    return plist, None

def discover_go_mod(args, localpath):
    plist = []
    files = lib_utils.find_files(localpath, 'go.mod')
    prop_dict = None
    for file_path in files:
        fp = lib_utils.tw_open(file_path, args.encoding)
        if fp == None:
            continue
        contents = fp.read()
        fp.close()
        if localpath.startswith('/tmp/'):
            file_path = file_path.replace(localpath+'/','')
        for line in contents.splitlines():
            line = line.strip()
            if line == '' or line.startswith('module ') or line.startswith('go ') or line.startswith('require (') or line.startswith(')') or line.startswith('replace') or line.startswith('exclude'):
                continue
            sp = line.split()
            if len(sp) < 2:
                continue
            ver = sp[1]
            if ver.startswith('v'):
                ver = ver[1:]
            prod = sp[0] + " " + ver + " source:"+file_path
            plist.append(prod)
    return plist, None

def discover_cargo_toml(args, localpath):
    plist = []
    files = lib_utils.find_files(localpath, 'Cargo.toml')
    prop_dict = None
    for file_path in files:
        fp = lib_utils.tw_open(file_path, args.encoding)
        if fp == None:
            continue
        contents = fp.read()
        tdict = toml.loads(contents)
        fp.close()
        if localpath.startswith('/tmp/'):
            file_path = file_path.replace(localpath+'/','')
        if 'dependencies' in tdict:
            for d in tdict['dependencies']:
                vers = str(tdict['dependencies'][d])
                if vers.startswith('{'):
                    if 'version' in tdict['dependencies'][d]:
                        ver = tdict['dependencies'][d]['version']
                    else:
                        continue
                else:
                    ver = tdict['dependencies'][d]
                prod = d + " " + ver
                prod = cleanse_semver_version(prod)
                prod = prod + " source:"+file_path
                plist.append(prod)
    return plist, None

def discover_pom_xml(args, localpath):
    plist = []
    files = lib_utils.find_files(localpath, 'pom.xml')
    prop_dict = None
    for file_path in files:
        fp = lib_utils.tw_open(file_path, args.encoding)
        if fp == None:
            continue
        contents = fp.read()
        xmldoc = None
        try:
            xmldoc = minidom.parseString(contents)
        except Exception:
            logging.error("Unable to parse pom.xml "+file_path)
            continue
        if prop_dict is None:
            prop_dict = { }
            curr_prop_dict = prop_dict
        else:
            curr_prop_dict = prop_dict.copy()

        version_elements = xmldoc.getElementsByTagName('version')
        for ve in version_elements:
            if ve.parentNode.nodeName == 'project':
                curr_prop_dict['project.version'] = ve.childNodes[0].data
                break

        prop_list = xmldoc.getElementsByTagName('properties')
        if len(prop_list) > 0:
            for p in prop_list:
                for item in p.childNodes:
                    if item.nodeType == item.TEXT_NODE or item.nodeType == item.COMMENT_NODE or len(item.childNodes) == 0:
                        continue
                    prop_name = item.nodeName
                    prop_value = item.childNodes[0].data
                    if prop_value.startswith('${'):
                        # Try to look-up value from prop_dict
                        lookup_prop = prop_value[2:-1]
                        prop_value = prop_dict.get(lookup_prop)
                    if prop_value is not None:
                        curr_prop_dict[prop_name] = prop_value

        dlist = xmldoc.getElementsByTagName('dependency')
        for d in dlist:
            gid = d.getElementsByTagName('groupId')
            if len(gid) == 0:
                gid = None
            else:
                gid = gid[0]
            aid = d.getElementsByTagName('artifactId')
            if len(aid) == 0:
                continue
            else:
                aid = aid[0]
            ver = d.getElementsByTagName('version')
            if len(ver) == 0:
                ver = None 
            else:
                ver = ver[0]
            libgname = ''
            if gid != None:
                libgname = gid.childNodes[0].data
            libname = aid.childNodes[0].data
            libver = ''
            if ver != None:
                libver = ver.childNodes[0].data
                if libver.startswith('${'):
                    prop_value = curr_prop_dict.get(libver[2:-1])
                    if prop_value is not None:
                        libver = prop_value
            if libgname == '':
                pname = libname + ' ' + libver
            else:
                pname = libgname + ':' + libname + ' ' + libver
            if localpath.startswith('/tmp/'):
                file_path = file_path.replace(localpath+'/','')
            pname = pname.strip() + " source:"+file_path
            if pname not in plist:
                plist.append(pname)
    return plist, None

def discover_gradle(args, localpath):
    plist = []
    files = lib_utils.find_files(localpath, 'dependencies.gradle')
    for file_path in files:
        fp = lib_utils.tw_open(file_path, args.encoding)
        if fp == None:
            continue
        contents = fp.read()
        for l in contents.splitlines():
            if 'group:' in l and 'name:' in l and 'version:' in l:
                arr = l.split(',')
                gname = arr[0].split('group:')[1].strip()
                lname = arr[1].split('name:')[1].strip()
                ver = arr[2].split('version:')[1].strip()
                if localpath.startswith('/tmp/'):
                    file_path = file_path.replace(localpath+'/','')
                pname = gname + ':' + lname + ' ' + ver + " source:"+file_path
                pname = pname.replace("'","")
                if pname not in plist:
                    plist.append(pname)
    return plist, None

def process_package_json_files(files, args, localpath):
    level = args.level
    encoding = args.encoding
    plist = []
    p1list = [] # 1st level dependencies
    for file_path in files:
        fp = lib_utils.tw_open(file_path, encoding)
        if fp == None:
            continue
        contents = fp.read()
        contents = contents.strip()
        if len(contents) == 0:
            continue
        cjson = ''
        try:
            cjson = json.loads(contents)
        except Exception:
            logging.error("Unable to parse package.json contents")
            continue
        if 'name' in cjson and 'version' in cjson:
            if localpath.startswith('/tmp/'):
                file_path = file_path.replace(localpath+'/','')
            pname = cjson['name'] + ' ' + cjson['version'] + " source:"+file_path
            if pname not in plist:
                plist.append(pname)
                p1list.append(pname)
        if 'dependencies' in cjson:
            ddict = cjson['dependencies']
            for d in ddict:
                content = ddict[d]
                if isinstance(content, dict):
                    ver = content['version']
                    if localpath.startswith('/tmp/'):
                        file_path = file_path.replace(localpath+'/','')
                    pname = d + ' ' + ver + " source:"+file_path
                    if pname not in plist:
                        plist.append(pname)
                        p1list.append(pname)
                    req_dict = content.get('requires')
                    if req_dict is None or level == 'shallow':
                        continue
                    for req_pname in req_dict:
                        if localpath.startswith('/tmp/'):
                            file_path = file_path.replace(localpath+'/','')
                        pname = req_pname + ' ' + req_dict[req_pname]
                        pname = cleanse_semver_version(pname) + " source:"+file_path
                        if pname not in plist:
                            plist.append(pname)
                else:
                    if localpath.startswith('/tmp/'):
                        file_path = file_path.replace(localpath+'/','')
                    pname = d + ' ' + ddict[d] + " source:"+file_path
                    pname = cleanse_semver_version(pname)
                    if pname not in plist:
                        plist.append(pname)
                        p1list.append(pname)
        if 'devDependencies' in cjson:
            ddict = cjson['devDependencies']
            for d in ddict:
                pname = d + ' ' + ddict[d]
                if localpath.startswith('/tmp/'):
                    file_path = file_path.replace(localpath+'/','')
                pname = cleanse_semver_version(pname) + " source:"+file_path
                if pname not in plist:
                    plist.append(pname)
                    p1list.append(pname)
        if 'optionalDependencies' in cjson:
            ddict = cjson['optionalDependencies']
            for d in ddict:
                pname = d + ' ' + ddict[d]
                if localpath.startswith('/tmp/'):
                    file_path = file_path.replace(localpath+'/','')
                pname = cleanse_semver_version(pname) + " source:"+file_path
                if pname not in plist:
                    plist.append(pname)
                    p1list.append(pname)
    return plist, p1list 

def filter_used_npm_dependencies(args, deplist, localpath):
    logging.debug("Number of dependencies before used filter: "+str(len(deplist)))
    dev_null_device = open(os.devnull, "w")
    fdlist = []
    f_handle, f_path = tempfile.mkstemp('.txt', 'tw-find-output-')
    find_cmd = "find "+localpath+" -type f -name '*.js' -or -name '*.ts' > " + f_path
    try:
        out = subprocess.check_output([find_cmd], stderr=dev_null_device, shell=True)
        out = out.decode(args.encoding)
    except subprocess.CalledProcessError:
        if logging_enabled:
            logging.debug("Error running command...unable to filter used npm depdencies")
        return deplist
    for d in deplist:
        dname = d.split()[0].strip()
        if dname in fdlist:
            continue
        #print("Checking dependency for "+d)
        cmd = "cat %s | xargs -r egrep -ni '(import|require|loader|plugins|%s).*['\"](%s|.?\d+)[\"']' -m 1 | wc -l " % (f_path, dname, dname)
        #print(cmd)
        try:
            out = subprocess.check_output([cmd], stderr=dev_null_device, shell=True)
            out = out.decode(args.encoding)
        except subprocess.CalledProcessError:
            if logging_enabled:
                logging.debug("Error running command")
            dev_null_device.close()
            continue
        if out.strip() != '0':
            #print(dname+" is used")
            fdlist.append(d)
    dev_null_device.close()
    os.remove(f_path)
    logging.debug("Number of dependencies after used filter: "+str(len(fdlist)))
    return fdlist

def filter_used_dotnet_dependencies(args, deplist, localpath):
    logging.debug("Number of dependencies before used filter: "+str(len(deplist)))
    dev_null_device = open(os.devnull, "w")
    fdlist = []
    f_handle, f_path = tempfile.mkstemp('.txt', 'tw-find-output-')
    find_cmd = "find "+localpath+" -type f -name '*.cs' > " + f_path
    try:
        out = subprocess.check_output([find_cmd], stderr=dev_null_device, shell=True)
        out = out.decode(args.encoding)
    except subprocess.CalledProcessError:
        if logging_enabled:
            logging.debug("Error running command...unable to filter used npm depdencies")
        return deplist
    for d in deplist:
        dname = d.split()[0].strip()
        if dname in fdlist:
            continue
        #logging.debug("Checking dependency for "+d)
        cmd = "cat %s | xargs -r egrep -ni 'using.*%s' -m 1 | wc -l " % (f_path, dname)
        try:
            out = subprocess.check_output([cmd], stderr=dev_null_device, shell=True)
            out = out.decode(args.encoding)
        except subprocess.CalledProcessError:
            if logging_enabled:
                logging.debug("Error running command")
            dev_null_device.close()
            continue
        if out.strip() != '0':
            #logging.debug(dname+" is used")
            fdlist.append(d)
    dev_null_device.close()
    os.remove(f_path)
    logging.debug("Number of dependencies after used filter: "+str(len(fdlist)))
    return fdlist

def discover_package_json(args, localpath):
    files = lib_utils.find_files(localpath, 'package-lock.json')
    if len(files) > 0:
        plist, p1list = process_package_json_files(files, args, localpath)
    else:
        files = lib_utils.find_files(localpath, 'package.json')
        plist, p1list = process_package_json_files(files, args, localpath)
    if len(plist) > 0:
        if args.include_unused_dependencies == False:
            logging.debug("Filtering out unused npm dependencies. This may take some time...")
            plist = filter_used_npm_dependencies(args, plist, localpath)
        else:
            logging.warning("Including unused dependencies")
            logging.warning("May increase false positives")
    return plist, p1list

def discover_packages_config(args, localpath):
    plist = []
    # Give first preference for dependencies from .csproj files
    verprops = {}
    propfiles = lib_utils.find_files(localpath, '.props')
    for propfile in propfiles:
        fp = open(propfile, mode='r')
        if fp == None:
            continue
        contents = fp.read()
        xmldoc = None
        try:
            xmldoc = minidom.parseString(contents)
        except Exception:
            traceback.print_exc()
            logging.error("Unable to parse propfile file: %s", propfile)
            continue
        pglist = xmldoc.getElementsByTagName('PropertyGroup')
        for pg in pglist:
            props = pg.childNodes
            for p in props:
                if p.nodeType != p.TEXT_NODE: 
                    if p.childNodes and len(p.childNodes) == 1 and p.tagName.endswith('Version'):
                        verprops[p.tagName] = p.childNodes[0].data
    files = lib_utils.find_files(localpath, '.csproj')
    for file_path in files:
        fp = open(file_path, mode='r')
        if fp == None:
            continue
        contents = fp.read()
        xmldoc = None
        try:
            xmldoc = minidom.parseString(contents)
        except Exception:
            traceback.print_exc()
            logging.error("Unable to parse csproj file: %s", file_path)
            continue
        temp_plist = xmldoc.getElementsByTagName('PackageReference')
        for p in temp_plist:
            libname = p.getAttribute('Include')
            if libname is None or len(libname.strip()) == 0:
                continue
            libver = p.getAttribute('Version')
            if localpath.startswith('/tmp/'):
                file_path = file_path.replace(localpath+'/','')
            if libver is not None and len(libver.strip()) > 0 and not libver.startswith('$('):
                if '*' in libver:
                    libver = libver.replace('.*', '.0')
                if ',' in libver:
                    libver = libver.replace('(', '')
                    libver = libver.replace(')', '')
                    libver = libver.replace('[', '')
                    libver = libver.replace(']', '')
                    tokens = libver.split(',')
                    if len(tokens[0].strip()) > 0:
                        libver = tokens[0].strip()
                    elif len(tokens[1].strip()) > 0:
                        libver = tokens[1].strip()
                    else:
                        libver = '' # this case should never occur
                pname = libname + ' ' + libver + " source:"+file_path
            elif libver is not None and libver.startswith('$('):
                verlabel = libver.replace('$(','').replace(')','')
                if verlabel in verprops:
                    libver = verprops[verlabel]
                    pname = libname + ' ' + libver + " source:"+file_path
                else:
                    pname = libname + " source:"+file_path
            else:
                pname = libname + " source:"+file_path
            if pname not in plist:
                plist.append(pname)

    if len(plist) > 0:
        if args.include_unused_dependencies == False:
            logging.debug("Filtering out unused dotnet dependencies. This may take some time...")
            plist = filter_used_dotnet_dependencies(args, plist, localpath)
        else:
            logging.warning("Including unused dependencies")
            logging.warning("May increase false positives")

    if len(plist) > 0:
        return plist, plist

    # Fallback to packages.config file if no dependencies found from .csproj files
    files = lib_utils.find_files(localpath, 'packages.config')
    for file_path in files:
        fp = lib_utils.tw_open(file_path, args.encoding)
        if fp == None:
            continue
        contents = fp.read()
        xmldoc = None
        try:
            xmldoc = minidom.parseString(contents)
        except Exception:
            logging.error("Unable to parse package config: %s", file_path)
            continue
        temp_plist = xmldoc.getElementsByTagName('package')
        for p in temp_plist:
            libname = p.getAttribute('id')
            libver = p.getAttribute('version')
            if localpath.startswith('/tmp/'):
                file_path = file_path.replace(localpath+'/','')
            pname = libname + ' ' + libver + " source:"+file_path
            if pname not in plist:
                plist.append(pname)

    if len(plist) > 0:
        if args.include_unused_dependencies == False:
            logging.debug("Filtering out unused nuget dependencies. This may take some time...")
            plist = filter_used_dotnet_dependencies(args, plist, localpath)
        else:
            logging.warning("Including unused dependencies")
            logging.warning("May increase false positives")

    return plist, plist

def discover_yarn(args, localpath):
    plist = []
    p1list = []
    files = lib_utils.find_files(localpath, 'yarn.lock')
    if len(files) == 0 and args.type is not None:
        files = lib_utils.find_files(localpath, 'package.json')
        plist = process_package_json_files(files, args, localpath)
        return plist

    for file_path in files:
        fp = lib_utils.tw_open(file_path, args.encoding)
        if fp == None:
            continue
        contents = fp.read()
        cline = contents.splitlines()
        dparse = False
        for index, l in enumerate(cline):
            if l.endswith(':') and 'dependencies' not in l.lower():
                dparse = False
                if l.startswith('"'):
                    l = l[1:]
                if l.startswith('@'):
                    libname = '@' + l.split('@')[1]
                else:
                    libname = l.split('@')[0]
                vline = cline[index+1]
                libver = vline.split()[1].replace('"','')
                if localpath.startswith('/tmp/'):
                    file_path = file_path.replace(localpath+'/','')
                pname = libname+' '+libver
                pname = cleanse_semver_version(pname) + " source:"+file_path
                if pname not in plist:
                    plist.append(pname)
                    p1list.append(pname)
            if l.endswith(':') and 'dependencies' in l.lower():
                dparse = True
                continue
            if dparse:
                pname = l.strip()
                if pname == '':
                    dparse = False
                    continue
                if localpath.startswith('/tmp/'):
                    file_path = file_path.replace(localpath+'/','')
                pname = cleanse_semver_version(pname) + " source:"+file_path
                if pname not in plist and args.type == 'deep':
                    plist.append(pname)
    return plist, p1list

def discover_ruby(args, localpath):
    pset = set() 
    p1list = [] # 1st level dependencies
    files = lib_utils.find_files(localpath, 'gemfile.lock')
    if len(files) == 0:
        files = lib_utils.find_files(localpath, 'Gemfile.lock')
    for file_path in files:
        fp = lib_utils.tw_open(file_path, args.encoding)
        if fp == None:
            continue
        contents = fp.read()
        cline = contents.splitlines()
        specsfound = False
        for index, l in enumerate(cline):
            raw_line = l
            l = l.strip()
            if l.startswith('specs:'):
                specsfound = True
                first_line_indent = -1
                continue
            if l == '':
                specsfound = False
                continue
            if specsfound:
                current_line_indent = lib_utils.get_indent(raw_line)
                if first_line_indent == -1:
                    first_line_indent = current_line_indent
                elif current_line_indent > first_line_indent and args.level == 'shallow':
                    # skip 2nd level dependencies
                    continue
                ls = l.split()
                gname = ls[0]
                gver = ''
                if len(ls) > 1:
                    for i in range(1,len(ls)):
                        gver = ls[i]
                        gver = re.findall(r'([0-9]+[0-9a-z]*(\.[0-9a-z]+)+)', gver)
                        if gver:
                            gver = gver[0][0]
                            break
                        else:
                            gver = ''
                pname = gname + ' ' + gver
                if localpath.startswith('/tmp/'):
                    file_path = file_path.replace(localpath+'/','')
                pname = pname.strip() + " source:"+file_path
                if pname not in pset:
                    pset.add(pname)
                    p1list.append(pname)
    return list(pset) , p1list

def discover_python(args, localpath):
    plist = []
    files = lib_utils.find_files(localpath, 'requirements.txt')
    for file_path in files:
        fp = lib_utils.tw_open(file_path, args.encoding)
        if fp == None:
            continue
        req = requirements.parse(fp)
        try:
            for r in req:
                prod = r.name
                if len(r.specs) > 0:
                    if localpath.startswith('/tmp/'):
                        file_path = file_path.replace(localpath+'/','')
                    prod = prod + ' ' + r.specs[0][1] + " source:"+file_path
                    if prod not in plist:
                        plist.append(prod)
        except:
            logging.error("Unable to parse python dependencies")
            continue
    """
    files = lib_utils.find_files(localpath, 'setup.py')
    for file_path in files:
        #Parse setup.py and return args and keywords args to its setup
        #function call

        mock_setup = textwrap.dedent('''\
        def setup(*args, **kwargs):
            __setup_calls__.append((args, kwargs))
        ''')
        parsed_mock_setup = ast.parse(mock_setup, filename=file_path)
        with open(file_path, 'rt') as setup_file:
            parsed = ast.parse(setup_file.read())
            for index, node in enumerate(parsed.body[:]):
                if (
                    not isinstance(node, ast.Expr) or
                    not isinstance(node.value, ast.Call)
                    #not isinstance(node.value, ast.Call) or
                    #node.value.func.id != 'setup'
                ):
                    continue
                parsed.body[index:index] = parsed_mock_setup.body
                break

        fixed = ast.fix_missing_locations(parsed)
        codeobj = compile(fixed, file_path, 'exec')
        local_vars = {}
        global_vars = {'__setup_calls__': []}
        cwd = os.getcwd()
        os.chdir(os.path.dirname(file_path))
        try:
            exec(codeobj, global_vars, local_vars)
        except Exception:
            # move on
            logging.warning("Unable to parse python dependencies %s", file_path)
            os.chdir(cwd)
            continue
        os.chdir(cwd)
        reqs = global_vars['__setup_calls__'][0][1]['install_requires']
        for r in reqs:
            prod = r
            prod = prod.replace("==", " ")
            prod = prod.replace(">=", " ")
            prod = prod.replace("<=", " ")
            prod = prod.replace("~=", " ")
            prod = prod.replace(">", " ")
            prod = prod.replace("<", " ")
            prod = prod + " source:" + file_path
            if prod not in plist:
                plist.append(prod)
    """

    files = lib_utils.find_files(localpath, 'pyproject.toml')
    for file_path in files:
        fp = lib_utils.tw_open(file_path, args.encoding)
        if fp == None:
            continue
        contents = fp.read()
        tdict = toml.loads(contents)
        fp.close()
        if localpath.startswith('/tmp/'):
            file_path = file_path.replace(localpath+'/','')
        if 'tool' in tdict and 'poetry' in tdict['tool'] and 'dependencies' in tdict['tool']['poetry']:
            for d in tdict['tool']['poetry']['dependencies']:
                ver = tdict['tool']['poetry']['dependencies'][d]
                if isinstance(ver, dict):
                    ver = ver['version']
                ver = cleanse_semver_version(ver)
                prod = d + " " + ver + " source:" + file_path
                if prod not in plist:
                    plist.append(prod)
        if 'project' in tdict and 'dependencies' in tdict['project']: 
            for d in tdict['project']['dependencies']:
                prod = d 
                prod = prod.replace("==", " ")
                prod = prod.replace(">=", " ")
                prod = prod.replace("<=", " ")
                prod = prod.replace("~=", " ")
                prod = prod.replace(">", " ")
                prod = prod.replace("<", " ")
                prod = prod + " source:" + file_path
                if prod not in plist:
                    plist.append(prod)
    return plist, None

def LOWORD(dword):
    return dword & 0x0000ffff

def HIWORD(dword): 
    return dword >> 16

def get_dll_version(path):
    try:
        pe = pefile.PE(path)
        if hasattr(pe, 'VS_FIXEDFILEINFO'):
            ms = pe.VS_FIXEDFILEINFO[0].FileVersionMS
            ls = pe.VS_FIXEDFILEINFO[0].FileVersionLS
            return "%d.%d.%d.%d" % (HIWORD (ms), LOWORD (ms), HIWORD (ls), LOWORD (ls))
        else:
            return None
    except Exception as e:
        logging.error("Unable to parse DLL file. Skipping.")
        return None

def discover_dll(args, localpath):
    plist = []
    files = lib_utils.find_files(localpath, '.dll')
    for file_path in files:
        dll_version = get_dll_version(file_path)
        if dll_version is None:
            continue
        if localpath.startswith('/tmp/'):
            file_path = file_path.replace(localpath+'/','')
        dll_details = os.path.basename(file_path) + " " + dll_version + " source:"+file_path

        plist.append(dll_details)
    return plist, None

def discover_jar(args, localpath):
    plist = []
    files = lib_utils.find_files(localpath, '.jar')
    for file_path in files:
        prod = ''
        ver = ''
        #print "Checking "+file_path
        try:
            zf = zipfile.ZipFile(file_path, 'r')
            try:
                metafile = zf.read('META-INF/MANIFEST.MF')
                if metafile:
                    for l in metafile.splitlines():
                        if l.startswith(b"Bundle-Version:"):
                            ver = l.split(b':')[1].strip()
                        if l.startswith(b"Bundle-Name:"):
                            prod = l.split(b':')[1].strip().lower().replace(b' ',b'-')
            except KeyError:
                #print "Error: No manifest found"
                pass
        except zipfile.BadZipfile:
            logging.warning("Unable to inspect file: %s", os.path.basename(file_path))

        if prod == '' or ver == '':
            jfile = os.path.basename(file_path)
            jfile = jfile.replace('.jar','')
            pattern = r'(?:(\d+\.(?:\d+\.)*\d+))'
            match = re.findall(pattern, jfile)
            if len(match) == 0:
                continue
            else:
                ver = match[0]
                prod = jfile.split(ver)[0][:-1]
        prod = str(prod) + ' ' + str(ver)
        if localpath.startswith('/tmp/'):
            file_path = file_path.replace(localpath+'/','')
        prod = prod.strip() + " source:"+file_path
        prod = prod.replace("b'","")
        prod = prod.replace("'","")
        prod = prod.replace('b"','')
        prod = prod.replace('"','')
        if prod != '':
            plist.append(prod)
    return plist, None

def discover_specified_type(repo_type, args, localpath):
    if repo_type not in SUPPORTED_TYPES:
        logging.error("Type not supported")
        return [], None

    plist = []
    p1list = []
    if repo_type == 'pip':
        plist, p1list = discover_python(args, localpath)
    elif repo_type == 'ruby':
        plist, p1list = discover_ruby(args, localpath)
    elif repo_type == 'yarn':
        plist, p1list = discover_yarn(args, localpath)
    elif repo_type == 'nuget':
        plist, p1list = discover_packages_config(args, localpath)
    elif repo_type == 'npm':
        plist, p1list = discover_package_json(args, localpath)
    elif repo_type == 'maven':
        plist, p1list = discover_pom_xml(args, localpath)
    elif repo_type == 'gradle':
        plist, p1list = discover_gradle(args, localpath)
    elif repo_type == 'dll':
        plist, p1list = discover_dll(args, localpath)
    elif repo_type == 'jar':
        plist, p1list = discover_jar(args, localpath)
    elif repo_type == 'cargo':
        plist, p1list = discover_cargo_toml(args, localpath)
    elif repo_type == 'go':
        plist, p1list = discover_go_mod(args, localpath)
    elif repo_type == 'composer':
        plist, p1list = discover_composer(args, localpath)

    return plist, p1list

def get_last_component(repo_path):
    if repo_path.startswith('http:') or repo_path.startswith('https:'):
        return repo_path.rsplit('/',1)[-1]
    else:
        return os.path.basename(os.path.normpath(repo_path))

def strip_source(plist):
    ret_list = []
    for pname in plist:
        index = pname.find(" source:")
        if index != -1:
            ret_list.append(pname[:index])
        else:
            ret_list.append(pname)
    return ret_list

def get_asset_id(args):
    asset_id = None
    if args.assetid == None or args.assetid.strip() == "":
        tokens = [args.handle.split('@')[0]]
        repo_path = args.repo
        if repo_path.startswith('http:') or repo_path.startswith('https:'):
            tokens.extend(repo_path.split('/')[3:])
            if args.branch is not None and args.branch.strip() != "":
                tokens.append(args.branch)
        else:
            tokens.append(os.path.basename(os.path.normpath(repo_path)))
        asset_id = "-".join(tokens)
    else:
        asset_id = args.assetid
    asset_id = asset_id.replace(' ','-')
    asset_id = asset_id.replace('/','-')
    asset_id = asset_id.replace(':','-')
    return asset_id

def discover_inventory(args, localpath):
    asset_name = None
    if args.assetname == None or args.assetname.strip() == "":
        asset_name = get_last_component(args.repo)
    else:
        asset_name = args.assetname

    atype = 'Source Repository'
    plist = []
    asset_tags = []
    tech2prod_dict = { }
    shallow_tech2prod_dict = { }
    if args.type is None:
        # If no type is specified, then process all supported types
        for repo_type in SUPPORTED_TYPES:
            temp_list, temp1list = discover_specified_type(repo_type, args, localpath)
            temp_list = list(set(temp_list))
            if temp_list is not None and len(temp_list) > 0:
                tech2prod_dict[repo_type] = strip_source(temp_list)
                if temp1list is not None and len(temp1list) > 0:
                    shallow_tech2prod_dict[repo_type] = strip_source(temp1list)
                plist.extend(temp_list)
                asset_tags.append(repo_type)
    else:
        plist, p1list = discover_specified_type(args.type, args, localpath)
        plist = list(set(plist))
        if plist is not None and len(plist) > 0:
            tech2prod_dict[args.type] = strip_source(plist)
            if p1list is not None and len(p1list) > 0:
                shallow_tech2prod_dict[args.type] = strip_source(p1list)
        asset_tags.append(args.type)

    if plist == None or len(plist) == 0:
        logging.info("Unable to identify any source code dependencies")

    asset_data = {}
    asset_data['id'] = get_asset_id(args)
    asset_data['name'] = asset_name
    asset_data['type'] = atype
    asset_data['owner'] = args.handle
    asset_data['products'] = plist
    asset_data['tags'] = asset_tags
    if len(tech2prod_dict) > 0:
        asset_data['compliance_metadata'] = {"source_metadata": {"technology_products":tech2prod_dict, "shallow_technology_products":shallow_tech2prod_dict}}
   
    lib_utils.update_tool_run_record()
    return [ asset_data ]

# Note this error routine assumes that the file was read-only and hence could not be deleted
def on_rm_error( func, path, exc_info):
    os.chmod( path, stat.S_IWRITE )
    os.unlink( path )

# note this only handles actual repos and not orgs (i.e. all repos in an org)
def get_inventory_helper(args):
    path = None
    if args.repo.startswith('http'):
        if os.path.isfile(GIT_PATH) == False:
            logging.error("git executable does not exist at [%s]", GIT_PATH)
            return None
        dev_null_device = open(os.devnull, "w")
        path = tempfile.mkdtemp()
        base_path = path
        new_repo = None
        logging.info("Cloning repo locally...")
        out = ""
        try:
            if args.branch and args.branch != '':
                cmdarr = [GIT_PATH, 'clone', '--branch', args.branch, args.repo, path+'/.']
            else:
                cmdarr = [GIT_PATH, 'clone', args.repo, path+'/.']
            #out = subprocess.check_output(cmdarr, stderr=dev_null_device)
            out = subprocess.check_output(cmdarr)
        except Exception as e:
            logging.error(out)
            logging.error(str(e))
            logging.error(traceback.format_exc())
            logging.error('Error cloning repo locally')
            shutil.rmtree(path, onerror = on_rm_error)
            return None
    elif os.path.isdir(args.repo):
        path = args.repo
        base_path = os.path.abspath(path)
        if base_path == os.path.dirname(base_path):
            base_path = "" # handle directory contained in root directory
    else:
        logging.error('Not a valid repo')
        return None

    assets = discover_inventory(args, path)
    if args.secrets_scan:
        logging.info("Discovering secrets/sensitive information. This may take some time.")
        secret_records = lib_code_secrets.scan_for_secrets(args, path, base_path)
        assets[0]['secrets'] = secret_records
    lib_utils.update_tool_run_record()

    code_issues = []
    if args.sast:
        logging.info("Performing static analysis. This may take some time.")
        sast_records = sast.run_sast(args, path, base_path)
        code_issues.extend(sast_records)
    lib_utils.update_tool_run_record()

    if args.iac_checks:
        logging.info("Identifying infrastructure as code (IaC) issues. This may take some time.")
        iac_records = iac.run_iac_checks(args, path, base_path)
        code_issues.extend(iac_records)
    lib_utils.update_tool_run_record()

    if len(code_issues) > 0:
        assets[0]['sast'] = code_issues

    if args.repo.startswith('http'):
        shutil.rmtree(path, onerror = on_rm_error)

    if len(assets[0]['products']) == 0 and (assets[0].get('secrets') is None or len(assets[0]['secrets']) == 0) and (assets[0].get('sast') is None or len(assets[0]['sast']) == 0):
        logging.warning("Nothing to report")
        if args.create_empty_asset is None or not args.create_empty_asset:
            return [] # if there are no products nor secrets then no assets to report
    return assets

def get_inventory(args):
    if args.gh_user:
        owner = args.gh_user
        dev_null_device = open(os.devnull, "w")
        # check if 'gh' command is available
        try:
            cmdarr = ['which', 'gh']
            out = subprocess.check_output(cmdarr, stderr=dev_null_device)
        except subprocess.CalledProcessError as e:
            logging.error("[gh] command not found")
            return None
        # check if user is logged in 'gh'
        try:
            cmdarr = ['gh', 'auth', 'status']
            out = subprocess.check_output(cmdarr, stderr=dev_null_device)
        except subprocess.CalledProcessError as e:
            logging.error("Please login using [gh auth login] command to list repositories")
            logging.debug("[gh auth status] command returned exit code [%s]", e.returncode)
            return None
        # list repo for org/user
        try:
            cmdarr = ['gh', 'repo', 'list', owner, '--json', 'name,url', '--limit', '65535']
            out = subprocess.check_output(cmdarr, stderr=dev_null_device)
        except subprocess.CalledProcessError as e:
            logging.error("Unable to list repos for [%s]", owner)
            logging.debug("[gh repo list %s --json name,url] command returned %s", owner, e.output)
            logging.debug("[gh repo list %s --json name,url] command returned exit code [%s]", owner, e.returncode)
            return None
        repos = json.loads(out.strip())
        assets = []
        for repo in repos:
            logging.info("Discovering repo [%s] as an asset", repo['name'])
            args.repo = repo['url']
            logging.info("Repo URL: %s", args.repo)
            temp_assets = get_inventory_helper(args)
            assets.extend(temp_assets)
        return assets
    else:
        return get_inventory_helper(args)



import sys
import os
import subprocess
import logging
import json

from . import utils
from .gcp_cis_tool import gcp_cis_utils
from . import docker

def get_latest_tag(imagename):
    tcmd = "container images list-tags "+imagename+" --sort-by=~timestamp"
    t_json = gcp_cis_utils.run_gcloud_cmd(tcmd)
    if t_json:
        if len(t_json[0]['tags']) != 0:
            return ':' + t_json[0]['tags'][-1]
        else:
            return '@' + t_json[0]['digest']
    return None

def get_digest(imagename):
    tcmd = "container images describe "+imagename
    t_json = gcp_cis_utils.run_gcloud_cmd(tcmd)
    if t_json:
        return t_json['image_summary']['digest']

def get_inventory(args):
    allassets = [] 
    if args.repository is None and args.image is None:
        logging.error("Either fully qualified image name (with repository and tag / digest) or repository url needs to be specified")
        return None
    gcp_cis_utils.set_encoding(args.encoding)
    if not args.image:
        ilist_cmd = "container images list --repository "+args.repository
        i_json = gcp_cis_utils.run_gcloud_cmd(ilist_cmd)
        logging.info("Found %d images in %s", len(i_json), args.repository)
        for i in i_json:
            tag = get_latest_tag(i['name'])
            if tag == None:
                logging.error("Unable to determine latest tag / digest for image. Skipping "+i['name'])
                continue
            logging.info("Using tag/digest '"+tag[1:]+"'")
            args.image = i['name'] + tag
            args.assetid = i['name'] + tag
            args.assetid = args.assetid.replace('/','-')
            args.assetid = args.assetid.replace(':','-')
            args.assetname = i['name'] + tag
            logging.info("Discovering image "+args.image)
            assets = docker.get_inventory(args, get_digest(args.image))
            if assets:
                allassets = allassets + assets
        for a in allassets:
            a['tags'].append('GCR')
        return allassets
    else:
        image = args.image.split('/')[-1]
        if ':' not in image and '@' not in image:
            tag = get_latest_tag(args.image)
            if tag == None:
                logging.error("Unable to determine latest tag / digest for image")
                return None 
            logging.info("Using tag/digest '"+tag[1:]+"'")
            args.image = args.image + tag
        args.assetid = args.image
        args.assetid = args.assetid.replace('/','-')
        args.assetid = args.assetid.replace(':','-')
        args.assetname = args.image
        logging.info("Discovering image "+args.image)
        assets = docker.get_inventory(args, get_digest(args.image))
        if assets != None:
            for a in assets:
                a['tags'].append('GCR')
        return assets 
